#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import os
import tempfile
from unittest import mock
from unittest import skip

from oslo_config import cfg
from oslo_utils import uuidutils

from ironic.common import exception
from ironic.common import neutron as neutron_common
from ironic.common import states
from ironic.conductor import task_manager
from ironic.drivers.modules.network import common
from ironic.objects import portgroup
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils

CONF = cfg.CONF


class TestCommonFunctions(db_base.DbTestCase):

    def setUp(self):
        super(TestCommonFunctions, self).setUp()
        self.node = obj_utils.create_test_node(self.context,
                                               network_interface='neutron')
        self.port = obj_utils.create_test_port(
            self.context, node_id=self.node.id, address='52:54:00:cf:2d:32')
        self.vif_id = "fake_vif_id"
        self.client = mock.MagicMock()

    def _objects_setup(self, set_physnets):
        # This portgroup contains 2 ports, both of them without VIF. The ports
        # are assigned to physnet physnet1.
        physical_network = 'physnet1' if set_physnets else None
        pg1 = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id,
            physical_network=physical_network)
        pg1_ports = []
        for i in range(2):
            pg1_ports.append(obj_utils.create_test_port(
                self.context, node_id=self.node.id,
                address='52:54:00:cf:2d:0%d' % i,
                physical_network=physical_network,
                uuid=uuidutils.generate_uuid(), portgroup_id=pg1.id))
        # This portgroup contains 3 ports, one of them with 'some-vif'
        # attached, so the two free ones should be considered standalone.
        # The ports are assigned physnet physnet2.
        physical_network = 'physnet2' if set_physnets else None
        pg2 = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id, address='00:54:00:cf:2d:04',
            name='foo2', physical_network=physical_network,
            uuid=uuidutils.generate_uuid())
        pg2_ports = []
        for i in range(2, 4):
            pg2_ports.append(obj_utils.create_test_port(
                self.context, node_id=self.node.id,
                address='52:54:00:cf:2d:0%d' % i,
                physical_network=physical_network,
                uuid=uuidutils.generate_uuid(), portgroup_id=pg2.id))
        pg2_ports.append(obj_utils.create_test_port(
            self.context, node_id=self.node.id,
            address='52:54:00:cf:2d:04',
            physical_network=physical_network,
            internal_info={common.NetType.TENANT.vif_key: 'some-vif'},
            uuid=uuidutils.generate_uuid(), portgroup_id=pg2.id))
        # This portgroup has 'some-vif-2' attached to it and contains one port,
        # so neither portgroup nor port can be considered free. The ports are
        # assigned physnet3.
        physical_network = 'physnet3' if set_physnets else None
        pg3 = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id, address='00:54:00:cf:2d:05',
            name='foo3', physical_network=physical_network,
            uuid=uuidutils.generate_uuid(),
            internal_info={common.NetType.TENANT.vif_key: 'some-vif-2'})
        pg3_ports = [obj_utils.create_test_port(
            self.context, node_id=self.node.id,
            address='52:54:00:cf:2d:05', uuid=uuidutils.generate_uuid(),
            physical_network=physical_network,
            portgroup_id=pg3.id)]
        return pg1, pg1_ports, pg2, pg2_ports, pg3, pg3_ports

    def test__get_free_portgroups_and_ports_no_port_physnets(self):
        self.node.network_interface = 'flat'
        self.node.save()
        pg1, pg1_ports, pg2, pg2_ports, pg3, pg3_ports = self._objects_setup(
            set_physnets=False)
        with task_manager.acquire(self.context, self.node.id) as task:
            free_port_like_objs = (
                common._get_free_portgroups_and_ports(task, self.vif_id,
                                                      {'anyphysnet'}))
        self.assertCountEqual(
            [pg1.uuid, self.port.uuid] + [p.uuid for p in pg2_ports[:2]],
            [p.uuid for p in free_port_like_objs])

    def test__get_free_portgroups_and_ports_no_physnets(self):
        self.node.network_interface = 'flat'
        self.node.save()
        pg1, pg1_ports, pg2, pg2_ports, pg3, pg3_ports = self._objects_setup(
            set_physnets=True)
        with task_manager.acquire(self.context, self.node.id) as task:
            free_port_like_objs = (
                common._get_free_portgroups_and_ports(task, self.vif_id,
                                                      set()))
        self.assertCountEqual(
            [pg1.uuid, self.port.uuid] + [p.uuid for p in pg2_ports[:2]],
            [p.uuid for p in free_port_like_objs])

    def test__get_free_portgroups_and_ports_no_matching_physnet(self):
        self.node.network_interface = 'flat'
        self.node.save()
        pg1, pg1_ports, pg2, pg2_ports, pg3, pg3_ports = self._objects_setup(
            set_physnets=True)
        with task_manager.acquire(self.context, self.node.id) as task:
            free_port_like_objs = (
                common._get_free_portgroups_and_ports(task, self.vif_id,
                                                      {'notaphysnet'}))
        self.assertCountEqual(
            [self.port.uuid],
            [p.uuid for p in free_port_like_objs])

    def test__get_free_portgroups_and_ports_physnet1(self):
        self.node.network_interface = 'flat'
        self.node.save()
        pg1, pg1_ports, pg2, pg2_ports, pg3, pg3_ports = self._objects_setup(
            set_physnets=True)
        with task_manager.acquire(self.context, self.node.id) as task:
            free_port_like_objs = (
                common._get_free_portgroups_and_ports(task, self.vif_id,
                                                      {'physnet1'}))
        self.assertCountEqual(
            [pg1.uuid, self.port.uuid],
            [p.uuid for p in free_port_like_objs])

    def test__get_free_portgroups_and_ports_physnet2(self):
        self.node.network_interface = 'flat'
        self.node.save()
        pg1, pg1_ports, pg2, pg2_ports, pg3, pg3_ports = self._objects_setup(
            set_physnets=True)
        with task_manager.acquire(self.context, self.node.id) as task:
            free_port_like_objs = (
                common._get_free_portgroups_and_ports(task, self.vif_id,
                                                      {'physnet2'}))
        self.assertCountEqual(
            [self.port.uuid] + [p.uuid for p in pg2_ports[:2]],
            [p.uuid for p in free_port_like_objs])

    def test__get_free_portgroups_and_ports_physnet3(self):
        self.node.network_interface = 'flat'
        self.node.save()
        pg1, pg1_ports, pg2, pg2_ports, pg3, pg3_ports = self._objects_setup(
            set_physnets=True)
        with task_manager.acquire(self.context, self.node.id) as task:
            free_port_like_objs = (
                common._get_free_portgroups_and_ports(task, self.vif_id,
                                                      {'physnet3'}))
        self.assertCountEqual(
            [self.port.uuid], [p.uuid for p in free_port_like_objs])

    def test__get_free_portgroups_and_ports_all_physnets(self):
        self.node.network_interface = 'flat'
        self.node.save()
        pg1, pg1_ports, pg2, pg2_ports, pg3, pg3_ports = self._objects_setup(
            set_physnets=True)
        physnets = {'physnet1', 'physnet2', 'physnet3'}
        with task_manager.acquire(self.context, self.node.id) as task:
            free_port_like_objs = (
                common._get_free_portgroups_and_ports(task, self.vif_id,
                                                      physnets))
        self.assertCountEqual(
            [pg1.uuid, self.port.uuid] + [p.uuid for p in pg2_ports[:2]],
            [p.uuid for p in free_port_like_objs])

    @mock.patch.object(neutron_common, 'validate_port_info', autospec=True)
    def test__get_free_portgroups_and_ports_neutron_missed(self, vpi_mock):
        vpi_mock.return_value = False
        with task_manager.acquire(self.context, self.node.id) as task:
            free_port_like_objs = (
                common._get_free_portgroups_and_ports(task, self.vif_id,
                                                      {'anyphysnet'}))
        self.assertCountEqual([], free_port_like_objs)

    @mock.patch.object(neutron_common, 'validate_port_info', autospec=True)
    def test__get_free_portgroups_and_ports_neutron(self, vpi_mock):
        vpi_mock.return_value = True
        with task_manager.acquire(self.context, self.node.id) as task:
            free_port_like_objs = (
                common._get_free_portgroups_and_ports(task, self.vif_id,
                                                      {'anyphysnet'}))
        self.assertCountEqual(
            [self.port.uuid], [p.uuid for p in free_port_like_objs])

    @mock.patch.object(neutron_common, 'validate_port_info', autospec=True)
    def test__get_free_portgroups_and_ports_flat(self, vpi_mock):
        self.node.network_interface = 'flat'
        self.node.save()
        vpi_mock.return_value = True
        with task_manager.acquire(self.context, self.node.id) as task:
            free_port_like_objs = (
                common._get_free_portgroups_and_ports(task, self.vif_id,
                                                      {'anyphysnet'}))
        self.assertCountEqual(
            [self.port.uuid], [p.uuid for p in free_port_like_objs])

    def test__get_free_portgroups_and_ports_port_uuid(self):
        self.node.network_interface = 'flat'
        self.node.save()
        pg1, pg1_ports, pg2, pg2_ports, pg3, pg3_ports = self._objects_setup(
            set_physnets=False)
        with task_manager.acquire(self.context, self.node.id) as task:
            free_port_like_objs = (
                common._get_free_portgroups_and_ports(
                    task, self.vif_id, {}, {'port_uuid': self.port.uuid}))
        self.assertCountEqual(
            [self.port.uuid],
            [p.uuid for p in free_port_like_objs])

    def test__get_free_portgroups_and_ports_portgroup_uuid(self):
        self.node.network_interface = 'flat'
        self.node.save()
        pg1, pg1_ports, pg2, pg2_ports, pg3, pg3_ports = self._objects_setup(
            set_physnets=False)
        with task_manager.acquire(self.context, self.node.id) as task:
            free_port_like_objs = (
                common._get_free_portgroups_and_ports(
                    task, self.vif_id, {}, {'portgroup_uuid': pg1.uuid}))
        self.assertCountEqual(
            [pg1.uuid],
            [p.uuid for p in free_port_like_objs])

    def test__get_free_portgroups_and_ports_portgroup_uuid_attached_vifs(self):
        self.node.network_interface = 'flat'
        self.node.save()
        pg1, pg1_ports, pg2, pg2_ports, pg3, pg3_ports = self._objects_setup(
            set_physnets=False)
        with task_manager.acquire(self.context, self.node.id) as task:
            free_port_like_objs = (
                common._get_free_portgroups_and_ports(
                    task, self.vif_id, {}, {'portgroup_uuid': pg2.uuid}))
        self.assertCountEqual(
            [],
            [p.uuid for p in free_port_like_objs])

    def test__get_free_portgroups_and_ports_no_matching_uuid(self):
        self.node.network_interface = 'flat'
        self.node.save()
        pg1, pg1_ports, pg2, pg2_ports, pg3, pg3_ports = self._objects_setup(
            set_physnets=False)
        with task_manager.acquire(self.context, self.node.id) as task:
            free_port_like_objs = (
                common._get_free_portgroups_and_ports(
                    task, self.vif_id, {},
                    {'port_uuid': uuidutils.generate_uuid()}))
        self.assertCountEqual(
            [],
            [p.uuid for p in free_port_like_objs])

    @mock.patch.object(neutron_common, 'validate_port_info', autospec=True,
                       return_value=True)
    def test_get_free_port_like_object_ports(self, vpi_mock):
        with task_manager.acquire(self.context, self.node.id) as task:
            res = common.get_free_port_like_object(task, self.vif_id,
                                                   {'anyphysnet'})
            self.assertEqual(self.port.uuid, res.uuid)

    @mock.patch.object(neutron_common, 'validate_port_info', autospec=True,
                       return_value=True)
    def test_get_free_port_like_object_ports_pxe_enabled_first(self, vpi_mock):
        self.port.pxe_enabled = False
        self.port.save()
        other_port = obj_utils.create_test_port(
            self.context, node_id=self.node.id, address='52:54:00:cf:2d:33',
            uuid=uuidutils.generate_uuid())
        with task_manager.acquire(self.context, self.node.id) as task:
            res = common.get_free_port_like_object(task, self.vif_id,
                                                   {'anyphysnet'})
            self.assertEqual(other_port.uuid, res.uuid)

    @mock.patch.object(neutron_common, 'validate_port_info', autospec=True,
                       return_value=True)
    def test_get_free_port_like_object_ports_physnet_match_first(self,
                                                                 vpi_mock):
        self.port.pxe_enabled = False
        self.port.physical_network = 'physnet1'
        self.port.save()
        obj_utils.create_test_port(
            self.context, node_id=self.node.id, address='52:54:00:cf:2d:33',
            uuid=uuidutils.generate_uuid())
        with task_manager.acquire(self.context, self.node.id) as task:
            res = common.get_free_port_like_object(task, self.vif_id,
                                                   {'physnet1'})
            self.assertEqual(self.port.uuid, res.uuid)

    @mock.patch.object(neutron_common, 'validate_port_info', autospec=True,
                       return_value=True)
    def test_get_free_port_like_object_ports_physnet_match_first2(self,
                                                                  vpi_mock):
        self.port.pxe_enabled = False
        self.port.physical_network = 'physnet1'
        self.port.save()
        pg = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id)
        obj_utils.create_test_port(
            self.context, node_id=self.node.id, address='52:54:00:cf:2d:01',
            uuid=uuidutils.generate_uuid(), portgroup_id=pg.id)
        with task_manager.acquire(self.context, self.node.id) as task:
            res = common.get_free_port_like_object(task, self.vif_id,
                                                   {'physnet1'})
            self.assertEqual(self.port.uuid, res.uuid)

    @mock.patch.object(neutron_common, 'validate_port_info', autospec=True,
                       return_value=True)
    def test_get_free_port_like_object_portgroup_first(self, vpi_mock):
        pg = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id)
        obj_utils.create_test_port(
            self.context, node_id=self.node.id, address='52:54:00:cf:2d:01',
            uuid=uuidutils.generate_uuid(), portgroup_id=pg.id)
        with task_manager.acquire(self.context, self.node.id) as task:
            res = common.get_free_port_like_object(task, self.vif_id,
                                                   {'anyphysnet'})
            self.assertEqual(pg.uuid, res.uuid)

    @mock.patch.object(neutron_common, 'validate_port_info', autospec=True,
                       return_value=True)
    def test_get_free_port_like_object_portgroup_physnet_match_first(self,
                                                                     vpi_mock):
        pg1 = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id)
        obj_utils.create_test_port(
            self.context, node_id=self.node.id, address='52:54:00:cf:2d:01',
            uuid=uuidutils.generate_uuid(), portgroup_id=pg1.id)
        pg2 = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id, uuid=uuidutils.generate_uuid(),
            name='pg2', address='52:54:00:cf:2d:01')
        obj_utils.create_test_port(
            self.context, node_id=self.node.id, address='52:54:00:cf:2d:02',
            uuid=uuidutils.generate_uuid(), portgroup_id=pg2.id,
            physical_network='physnet1')
        with task_manager.acquire(self.context, self.node.id) as task:
            res = common.get_free_port_like_object(task, self.vif_id,
                                                   {'physnet1'})
            self.assertEqual(pg2.uuid, res.uuid)

    @mock.patch.object(neutron_common, 'validate_port_info', autospec=True,
                       return_value=True)
    def test_get_free_port_like_object_ignores_empty_portgroup(self, vpi_mock):
        obj_utils.create_test_portgroup(self.context, node_id=self.node.id)
        with task_manager.acquire(self.context, self.node.id) as task:
            res = common.get_free_port_like_object(task, self.vif_id,
                                                   {'anyphysnet'})
            self.assertEqual(self.port.uuid, res.uuid)

    @mock.patch.object(neutron_common, 'validate_port_info', autospec=True,
                       return_value=True)
    def test_get_free_port_like_object_ignores_standalone_portgroup(
            self, vpi_mock):
        self.port.destroy()
        pg = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id)
        obj_utils.create_test_port(
            self.context, node_id=self.node.id, address='52:54:00:cf:2d:01',
            uuid=uuidutils.generate_uuid(), portgroup_id=pg.id,
            internal_info={common.NetType.TENANT.vif_key: 'some-vif'})
        free_port = obj_utils.create_test_port(
            self.context, node_id=self.node.id, address='52:54:00:cf:2d:02',
            uuid=uuidutils.generate_uuid(), portgroup_id=pg.id)
        with task_manager.acquire(self.context, self.node.id) as task:
            res = common.get_free_port_like_object(task, self.vif_id,
                                                   {'anyphysnet'})
            self.assertEqual(free_port.uuid, res.uuid)

    @mock.patch.object(neutron_common, 'validate_port_info', autospec=True,
                       return_value=True)
    def test_get_free_port_like_object_vif_attached_to_portgroup(
            self, vpi_mock):
        pg = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id,
            internal_info={common.NetType.TENANT.vif_key: self.vif_id})
        obj_utils.create_test_port(
            self.context, node_id=self.node.id, address='52:54:00:cf:2d:01',
            uuid=uuidutils.generate_uuid(), portgroup_id=pg.id)
        with task_manager.acquire(self.context, self.node.id) as task:
            self.assertRaisesRegex(
                exception.VifAlreadyAttached,
                r"already attached to Ironic Portgroup",
                common.get_free_port_like_object,
                task, self.vif_id, {'anyphysnet'})

    @mock.patch.object(neutron_common, 'validate_port_info', autospec=True,
                       return_value=True)
    def test_get_free_port_like_object_vif_attached_to_port(self, vpi_mock):
        self.port.internal_info = {common.NetType.TENANT.vif_key: self.vif_id}
        self.port.save()
        with task_manager.acquire(self.context, self.node.id) as task:
            self.assertRaisesRegex(
                exception.VifAlreadyAttached,
                r"already attached to Ironic Port\b",
                common.get_free_port_like_object,
                task, self.vif_id, {'anyphysnet'})

    @mock.patch.object(neutron_common, 'validate_port_info', autospec=True,
                       return_value=True)
    def test_get_free_port_like_object_nothing_free(self, vpi_mock):
        self.port.internal_info = {
            common.NetType.TENANT.vif_key: 'another-vif'
        }
        self.port.save()
        with task_manager.acquire(self.context, self.node.id) as task:
            self.assertRaises(exception.NoFreePhysicalPorts,
                              common.get_free_port_like_object,
                              task, self.vif_id, {'anyphysnet'})

    @mock.patch.object(neutron_common, 'validate_port_info', autospec=True,
                       return_value=True)
    def test_get_free_port_like_object_no_matching_physnets(self, vpi_mock):
        self.port.physical_network = 'physnet1'
        self.port.save()
        with task_manager.acquire(self.context, self.node.id) as task:
            self.assertRaises(exception.NoFreePhysicalPorts,
                              common.get_free_port_like_object,
                              task, self.vif_id, {'physnet2'})

    @mock.patch.object(neutron_common, 'update_neutron_port', autospec=True)
    @mock.patch.object(neutron_common, 'wait_for_port_status', autospec=True)
    @mock.patch.object(neutron_common, 'get_client', autospec=True)
    def test_plug_port_to_tenant_network_client(self, mock_gc,
                                                wait_mock_status,
                                                mock_update):
        self.port.internal_info = {common.NetType.TENANT.vif_key: self.vif_id}
        self.port.save()
        with task_manager.acquire(self.context, self.node.id) as task:
            common.plug_port_to_tenant_network(task, self.port,
                                               client=mock.MagicMock())
        self.assertFalse(mock_gc.called)
        self.assertTrue(mock_update.called)

    @mock.patch.object(neutron_common, 'update_neutron_port', autospec=True)
    @mock.patch.object(neutron_common, 'wait_for_port_status', autospec=True)
    @mock.patch.object(neutron_common, 'get_client', autospec=True)
    def test_plug_port_to_tenant_network_no_client(self, mock_gc,
                                                   wait_mock_status,
                                                   mock_update):
        self.port.internal_info = {common.NetType.TENANT.vif_key: self.vif_id}
        self.port.save()
        with task_manager.acquire(self.context, self.node.id) as task:
            common.plug_port_to_tenant_network(task, self.port)
        self.assertTrue(mock_gc.called)
        self.assertTrue(mock_update.called)

    @mock.patch.object(neutron_common, 'get_client', autospec=True)
    def test_plug_port_to_tenant_network_no_tenant_vif(self, mock_gc):
        nclient = mock.MagicMock()
        mock_gc.return_value = nclient
        self.port.extra = {}
        self.port.save()
        with task_manager.acquire(self.context, self.node.id) as task:
            self.assertRaisesRegex(
                exception.VifNotAttached,
                "not associated with port %s" % self.port.uuid,
                common.plug_port_to_tenant_network,
                task, self.port)

    @mock.patch.object(neutron_common, 'wait_for_host_agent', autospec=True)
    @mock.patch.object(neutron_common, 'wait_for_port_status', autospec=True)
    @mock.patch.object(neutron_common, 'update_neutron_port', autospec=True)
    @mock.patch.object(neutron_common, 'get_client', autospec=True)
    def test_plug_port_to_tenant_network_smartnic_port(
            self, mock_gc, mock_update, wait_port_mock, wait_agent_mock):
        nclient = mock.MagicMock()
        mock_gc.return_value = nclient
        local_link_connection = self.port.local_link_connection
        local_link_connection['hostname'] = 'hostname'
        self.port.local_link_connection = local_link_connection
        self.port.internal_info = {common.NetType.TENANT.vif_key: self.vif_id}
        self.port.is_smartnic = True
        self.port.save()
        with task_manager.acquire(self.context, self.node.id) as task:
            common.plug_port_to_tenant_network(task, self.port)
            wait_agent_mock.assert_called_once_with(
                nclient, 'hostname')
            wait_port_mock.assert_called_once_with(
                nclient, self.vif_id, 'ACTIVE', fail_on_binding_failure=True)
            self.assertTrue(mock_update.called)

    @mock.patch.object(neutron_common, 'update_neutron_port', autospec=True)
    @mock.patch.object(neutron_common, 'wait_for_port_status', autospec=True)
    @mock.patch.object(neutron_common, 'get_client', autospec=True)
    def test_plug_port_to_tenant_network_with_physical_network(
            self, mock_gc, wait_mock_status, mock_update):
        # Test that physical_network is included in binding:profile for port
        nclient = mock.MagicMock()
        mock_gc.return_value = nclient
        self.port.internal_info = {common.NetType.TENANT.vif_key: self.vif_id}
        self.port.physical_network = 'physnet1'
        self.port.save()

        expected_attrs = {
            'binding:vnic_type': neutron_common.VNIC_BAREMETAL,
            'binding:host_id': self.node.uuid,
            'mac_address': self.port.address,
            'binding:profile': {
                'local_link_information': [self.port.local_link_connection],
                'physical_network': 'physnet1'
            }
        }

        with task_manager.acquire(self.context, self.node.id) as task:
            common.plug_port_to_tenant_network(task, self.port)
            mock_update.assert_called_once_with(
                task.context, self.vif_id, expected_attrs)

    @mock.patch.object(neutron_common, 'update_neutron_port', autospec=True)
    @mock.patch.object(neutron_common, 'wait_for_port_status', autospec=True)
    @mock.patch.object(neutron_common, 'get_client', autospec=True)
    def test_plug_portgroup_to_tenant_network_with_physical_network(
            self, mock_gc, wait_mock_status, mock_update):
        # Test that physical_network is included in binding:profile for
        # a portgroup
        nclient = mock.MagicMock()
        mock_gc.return_value = nclient
        pg = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id, address='00:54:00:cf:2d:01',
            physical_network='physnet1')
        port1 = obj_utils.create_test_port(
            self.context, node_id=self.node.id, address='52:54:00:cf:2d:01',
            portgroup_id=pg.id, uuid=uuidutils.generate_uuid())
        port2 = obj_utils.create_test_port(
            self.context, node_id=self.node.id, address='52:54:00:cf:2d:02',
            portgroup_id=pg.id, uuid=uuidutils.generate_uuid())
        pg.internal_info = {common.NetType.TENANT.vif_key: self.vif_id}
        pg.save()

        expected_attrs = {
            'binding:vnic_type': neutron_common.VNIC_BAREMETAL,
            'binding:host_id': self.node.uuid,
            'mac_address': pg.address,
            'binding:profile': {
                'local_link_information': [port1.local_link_connection,
                                           port2.local_link_connection],
                'local_group_information': {
                    'id': pg.uuid,
                    'name': pg.name,
                    'bond_mode': pg.mode,
                    'bond_properties': {}
                },
                'physical_network': 'physnet1'
            }
        }

        with task_manager.acquire(self.context, self.node.id) as task:
            common.plug_port_to_tenant_network(task, pg)
            mock_update.assert_called_once_with(
                task.context, self.vif_id, expected_attrs)


class TestVifPortIDMixin(db_base.DbTestCase):

    def setUp(self):
        super(TestVifPortIDMixin, self).setUp()
        self.interface = common.VIFPortIDMixin()
        self.node = obj_utils.create_test_node(self.context,
                                               network_interface='neutron')
        self.port = obj_utils.create_test_port(
            self.context, node_id=self.node.id,
            address='52:54:00:cf:2d:32',
            internal_info={
                common.NetType.TENANT.vif_key: uuidutils.generate_uuid()
            },
            extra={'client-id': 'fake1'})
        network_data_file = os.path.join(
            os.path.dirname(__file__), 'json_samples', 'network_data.json')
        with open(network_data_file, 'rb') as fl:
            self.network_data = json.load(fl)

    def test__save_vif_to_port_like_obj_port(self):
        self.port.extra = {}
        self.port.save()
        vif_id = "fake_vif_id"
        self.interface._save_vif_to_port_like_obj(self.port, vif_id)
        self.port.refresh()
        self.assertIn(common.NetType.TENANT.vif_key, self.port.internal_info)
        self.assertEqual(vif_id,
                         self.port.internal_info[common.NetType.TENANT.vif_key])
        self.assertEqual({}, self.port.extra)

    def test__save_vif_to_port_like_obj_portgroup(self):
        vif_id = "fake_vif_id"
        pg = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id)
        obj_utils.create_test_port(
            self.context, node_id=self.node.id, address='52:54:00:cf:2d:01',
            portgroup_id=pg.id, uuid=uuidutils.generate_uuid()
        )
        self.interface._save_vif_to_port_like_obj(pg, vif_id)
        pg.refresh()
        self.assertIn(common.NetType.TENANT.vif_key, pg.internal_info)
        self.assertEqual(vif_id,
                         pg.internal_info[common.NetType.TENANT.vif_key])
        self.assertEqual({}, pg.extra)

    def test__clear_vif_from_port_like_obj_in_extra_port(self):
        self.interface._clear_vif_from_port_like_obj(self.port)
        self.port.refresh()
        self.assertNotIn('vif_port_id', self.port.extra)
        self.assertNotIn(common.NetType.TENANT.vif_key,
                         self.port.internal_info)

    def test__clear_vif_from_port_like_obj_in_internal_info_port(self):
        self.port.internal_info = {
            common.NetType.TENANT.vif_key: self.port.internal_info[
                common.NetType.TENANT.vif_key
            ]
        }
        self.port.extra = {}
        self.port.save()

        self.interface._clear_vif_from_port_like_obj(self.port)
        self.port.refresh()
        self.assertNotIn('vif_port_id', self.port.extra)
        self.assertNotIn(common.NetType.TENANT.vif_key,
                         self.port.internal_info)

    def test__clear_vif_from_port_like_obj_in_extra_portgroup(self):
        vif_id = uuidutils.generate_uuid()
        pg = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id,
            extra={'vif_port_id': vif_id})
        obj_utils.create_test_port(
            self.context, node_id=self.node.id, address='52:54:00:cf:2d:01',
            portgroup_id=pg.id, uuid=uuidutils.generate_uuid()
        )
        self.interface._clear_vif_from_port_like_obj(pg)
        pg.refresh()
        self.assertNotIn('vif_port_id', pg.extra)
        self.assertNotIn(common.NetType.TENANT.vif_key, pg.internal_info)

    def test__clear_vif_from_port_like_obj_in_internal_info_portgroup(self):
        vif_id = uuidutils.generate_uuid()
        pg = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id,
            internal_info={common.NetType.TENANT.vif_key: vif_id})
        obj_utils.create_test_port(
            self.context, node_id=self.node.id, address='52:54:00:cf:2d:01',
            portgroup_id=pg.id, uuid=uuidutils.generate_uuid()
        )
        self.interface._clear_vif_from_port_like_obj(pg)
        pg.refresh()
        self.assertNotIn('vif_port_id', pg.extra)
        self.assertNotIn(common.NetType.TENANT.vif_key, pg.internal_info)

    def test__get_port_like_obj_by_vif_id_in_internal_info(self):
        vif_id = self.port.internal_info[common.NetType.TENANT.vif_key]
        self.port.internal_info = {common.NetType.TENANT.vif_key: vif_id}
        self.port.extra = {}
        self.port.save()
        with task_manager.acquire(self.context, self.node.id) as task:
            result = self.interface._get_port_like_obj_by_vif_id(task, vif_id)
        self.assertEqual(self.port.id, result.id)

    def test__get_port_like_obj_by_vif_id_not_attached(self):
        vif_id = self.port.internal_info[common.NetType.TENANT.vif_key]
        self.port.internal_info = {}
        self.port.save()
        with task_manager.acquire(self.context, self.node.id) as task:
            self.assertRaisesRegex(exception.VifNotAttached,
                                   "it is not attached to it.",
                                   self.interface._get_port_like_obj_by_vif_id,
                                   task, vif_id)

    def test__get_vif_id_by_port_like_obj_in_internal_info(self):
        vif_id = self.port.internal_info[common.NetType.TENANT.vif_key]
        self.port.internal_info = {common.NetType.TENANT.vif_key: vif_id}
        self.port.extra = {}
        self.port.save()
        result = self.interface._get_vif_id_by_port_like_obj(self.port)
        self.assertEqual(vif_id, result)

    def test__get_vif_id_by_port_like_obj_not_attached(self):
        self.port.internal_info = {}
        self.port.save()
        result = self.interface._get_vif_id_by_port_like_obj(self.port)
        self.assertIsNone(result)

    def test_vif_list_port_and_portgroup(self):
        vif_id = uuidutils.generate_uuid()
        self.port.internal_info = {common.NetType.TENANT.vif_key: vif_id}
        self.port.save()
        pg_vif_id = uuidutils.generate_uuid()
        portgroup = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id,
            address='52:54:00:00:00:00',
            internal_info={common.NetType.TENANT.vif_key: pg_vif_id})
        obj_utils.create_test_port(
            self.context, node_id=self.node.id, portgroup_id=portgroup.id,
            address='52:54:00:cf:2d:01', uuid=uuidutils.generate_uuid())
        with task_manager.acquire(self.context, self.node.id) as task:
            vifs = self.interface.vif_list(task)
            self.assertCountEqual([{'id': pg_vif_id}, {'id': vif_id}], vifs)

    def test_vif_list_internal(self):
        vif_id = uuidutils.generate_uuid()
        self.port.internal_info = {common.NetType.TENANT.vif_key: vif_id}
        self.port.save()
        pg_vif_id = uuidutils.generate_uuid()
        portgroup = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id,
            address='52:54:00:00:00:00',
            internal_info={common.NetType.TENANT.vif_key: pg_vif_id})
        obj_utils.create_test_port(
            self.context, node_id=self.node.id, portgroup_id=portgroup.id,
            address='52:54:00:cf:2d:01', uuid=uuidutils.generate_uuid())
        with task_manager.acquire(self.context, self.node.id) as task:
            vifs = self.interface.vif_list(task)
            self.assertCountEqual([{'id': pg_vif_id}, {'id': vif_id}], vifs)

    def test_vif_list_extra_and_internal_priority(self):
        # TODO(TheJulia): Remove in Xena?
        vif_id = uuidutils.generate_uuid()
        vif_id2 = uuidutils.generate_uuid()
        self.port.extra = {'vif_port_id': vif_id2}
        self.port.internal_info = {common.NetType.TENANT.vif_key: vif_id}
        self.port.save()
        with task_manager.acquire(self.context, self.node.id) as task:
            vifs = self.interface.vif_list(task)
            self.assertEqual([{'id': vif_id}], vifs)

    def test_get_current_vif_internal_info_cleaning(self):
        internal_info = {common.NetType.CLEANING.vif_key: 'foo',
                         common.NetType.TENANT.vif_key: 'bar'}
        self.port.internal_info = internal_info
        self.port.save()
        with task_manager.acquire(self.context, self.node.id) as task:
            vif = self.interface.get_current_vif(task, self.port)
            self.assertEqual('foo', vif)

    def test_get_current_vif_internal_info_provisioning(self):
        internal_info = {common.NetType.PROVISIONING.vif_key: 'foo',
                         common.NetType.TENANT.vif_key: 'bar'}
        self.port.internal_info = internal_info
        self.port.save()
        with task_manager.acquire(self.context, self.node.id) as task:
            vif = self.interface.get_current_vif(task, self.port)
            self.assertEqual('foo', vif)

    def test_get_current_vif_internal_info_tenant_vif(self):
        internal_info = {common.NetType.TENANT.vif_key: 'bar'}
        self.port.internal_info = internal_info
        self.port.save()
        with task_manager.acquire(self.context, self.node.id) as task:
            vif = self.interface.get_current_vif(task, self.port)
            self.assertEqual('bar', vif)

    def test_get_current_vif_internal_info_rescuing(self):
        internal_info = {common.NetType.RESCUING.vif_key: 'foo',
                         common.NetType.TENANT.vif_key: 'bar'}
        self.port.internal_info = internal_info
        self.port.save()
        with task_manager.acquire(self.context, self.node.id) as task:
            vif = self.interface.get_current_vif(task, self.port)
            self.assertEqual('foo', vif)

    def test_get_current_vif_internal_info_inspection(self):
        internal_info = {common.NetType.INSPECTION.vif_key: 'foo'}
        self.port.internal_info = internal_info
        self.port.save()
        with task_manager.acquire(self.context, self.node.id) as task:
            vif = self.interface.get_current_vif(task, self.port)
            self.assertEqual('foo', vif)

    def test_get_current_vif_internal_info_servicing(self):
        internal_info = {common.NetType.SERVICING.vif_key: 'foo',
                         common.NetType.TENANT.vif_key: 'bar'}
        self.port.internal_info = internal_info
        self.port.save()
        with task_manager.acquire(self.context, self.node.id) as task:
            vif = self.interface.get_current_vif(task, self.port)
            self.assertEqual('foo', vif)

    def test_get_current_vif_none(self):
        internal_info = extra = {}
        self.port.internal_info = internal_info
        self.port.extra = extra
        self.port.save()
        with task_manager.acquire(self.context, self.node.id) as task:
            vif = self.interface.get_current_vif(task, self.port)
            self.assertIsNone(vif)

    def test_get_current_vif_tenant_last(self):
        self.assertEqual(list(common.NetType.__members__.values())[-1],
                         common.NetType.TENANT)


class TestNeutronVifPortIDMixin(db_base.DbTestCase):

    def setUp(self):
        super(TestNeutronVifPortIDMixin, self).setUp()
        self.interface = common.NeutronVIFPortIDMixin()
        self.node = obj_utils.create_test_node(self.context,
                                               network_interface='neutron')
        self.port = obj_utils.create_test_port(
            self.context, node_id=self.node.id,
            address='52:54:00:cf:2d:32',
            internal_info={
                common.NetType.TENANT.vif_key: uuidutils.generate_uuid()},
            extra={'client-id': 'fake1'})
        self.neutron_port = {'id': '132f871f-eaec-4fed-9475-0d54465e0f00',
                             'mac_address': '52:54:00:cf:2d:32'}

    @mock.patch.object(neutron_common, 'update_neutron_port',
                       autospec=True)
    @mock.patch.object(common.VIFPortIDMixin, '_save_vif_to_port_like_obj',
                       autospec=True)
    @mock.patch.object(common, 'get_free_port_like_object', autospec=True)
    @mock.patch.object(neutron_common, 'get_client', autospec=True)
    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    @mock.patch.object(neutron_common, 'get_physnets_by_port_uuid',
                       autospec=True)
    def test_vif_attach(self, mock_gpbpi, mock_upa, mock_client, mock_gfp,
                        mock_save, mock_update_port):
        vif = {'id': "fake_vif_id"}
        mock_gfp.return_value = self.port
        mock_gpbpi.return_value = {'physnet1', 'physnet2'}
        with task_manager.acquire(self.context, self.node.id) as task:
            self.interface.vif_attach(task, vif)
            mock_client.assert_called_once_with(context=task.context)
            mock_upa.assert_called_once_with(
                "fake_vif_id", self.port.address, context=task.context)
            mock_update_port.assert_has_calls([
                mock.call(None, 'fake_vif_id',
                          {'binding:host_id': '', 'binding:profile': {}},
                          None),
                mock.call(None, 'fake_vif_id',
                          {'mac_address': None}, None),
                mock.call(task.context, mock.ANY,
                          {'binding:vnic_type': 'baremetal',
                           'binding:host_id': mock.ANY,
                           'binding:profile': {}},
                          client=None)
            ])
        mock_gpbpi.assert_called_once_with(mock_client.return_value,
                                           'fake_vif_id')
        mock_gfp.assert_called_once_with(task, 'fake_vif_id',
                                         mock_gpbpi.return_value,
                                         {'id': 'fake_vif_id'})
        mock_save.assert_called_once_with(self.port, "fake_vif_id")

    @mock.patch.object(common.VIFPortIDMixin, '_save_vif_to_port_like_obj',
                       autospec=True)
    @mock.patch.object(common, 'get_free_port_like_object', autospec=True)
    @mock.patch.object(neutron_common, 'get_client', autospec=True)
    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    @mock.patch.object(neutron_common, 'get_physnets_by_port_uuid',
                       autospec=True)
    def test_vif_attach_failure(self, mock_gpbpi, mock_upa, mock_client,
                                mock_gfp, mock_save):
        vif = {'id': "fake_vif_id"}
        mock_gfp.side_effect = exception.NoFreePhysicalPorts(vif='fake-vif')
        with task_manager.acquire(self.context, self.node.id) as task:
            self.assertRaises(exception.NoFreePhysicalPorts,
                              self.interface.vif_attach, task, vif)
        mock_gfp.assert_called_once_with(task, 'fake_vif_id',
                                         mock_gpbpi.return_value,
                                         {'id': 'fake_vif_id'})
        self.assertFalse(mock_save.called)

    @mock.patch.object(neutron_common, 'unbind_neutron_port_if_bound',
                       autospec=True)
    @mock.patch.object(neutron_common, 'update_neutron_port',
                       autospec=True)
    @mock.patch.object(common.VIFPortIDMixin, '_save_vif_to_port_like_obj',
                       autospec=True)
    @mock.patch.object(common, 'get_free_port_like_object', autospec=True)
    @mock.patch.object(neutron_common, 'get_client', autospec=True)
    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    @mock.patch.object(neutron_common, 'get_physnets_by_port_uuid',
                       autospec=True)
    def test_vif_attach_with_physnet(self, mock_gpbpi, mock_upa, mock_client,
                                     mock_gfp, mock_save, mock_port_update,
                                     mock_unbind):
        self.port.physical_network = 'physnet1'
        self.port.save()
        vif = {'id': "fake_vif_id"}
        mock_gpbpi.return_value = {'physnet1'}
        mock_gfp.return_value = self.port
        with task_manager.acquire(self.context, self.node.id) as task:
            self.interface.vif_attach(task, vif)
            mock_client.assert_called_once_with(context=task.context)
            mock_unbind.assert_called_once_with("fake_vif_id", client=mock.ANY,
                                                context=task.context)
            mock_upa.assert_called_once_with(
                "fake_vif_id", self.port.address, context=task.context)
            mock_port_update.assert_called_once_with(
                task.context, mock.ANY,
                {'binding:vnic_type': 'baremetal',
                 'binding:host_id': mock.ANY,
                 'binding:profile': {}},
                client=None)
        mock_gpbpi.assert_called_once_with(mock_client.return_value,
                                           'fake_vif_id')
        mock_gfp.assert_called_once_with(task, 'fake_vif_id', {'physnet1'},
                                         {'id': 'fake_vif_id'})
        mock_save.assert_called_once_with(self.port, "fake_vif_id")

    @mock.patch.object(neutron_common, 'update_neutron_port',
                       autospec=True)
    @mock.patch.object(common.VIFPortIDMixin, '_save_vif_to_port_like_obj',
                       autospec=True)
    @mock.patch.object(common, 'plug_port_to_tenant_network', autospec=True)
    @mock.patch.object(common, 'get_free_port_like_object', autospec=True)
    @mock.patch.object(neutron_common, 'get_client', autospec=True)
    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    @mock.patch.object(neutron_common, 'get_physnets_by_port_uuid',
                       autospec=True)
    def test_vif_attach_active_node(self, mock_gpbpi, mock_upa, mock_client,
                                    mock_gfp, mock_plug, mock_save,
                                    mock_port_update):
        self.node.provision_state = states.ACTIVE
        self.node.save()
        vif = {'id': "fake_vif_id"}
        mock_gfp.return_value = self.port
        mock_gpbpi.return_value = {'physnet1'}
        with task_manager.acquire(self.context, self.node.id) as task:
            self.interface.vif_attach(task, vif)
            mock_client.assert_called_once_with(context=task.context)
            mock_upa.assert_called_once_with(
                "fake_vif_id", self.port.address, context=task.context)
            mock_port_update.assert_not_called()
        mock_gpbpi.assert_called_once_with(mock_client.return_value,
                                           "fake_vif_id")
        mock_gfp.assert_called_once_with(task, 'fake_vif_id',
                                         set(['physnet1']),
                                         {'id': 'fake_vif_id'})
        mock_save.assert_called_once_with(self.port, "fake_vif_id")
        mock_plug.assert_called_once_with(task, self.port, mock.ANY)

    @mock.patch.object(common.VIFPortIDMixin, '_save_vif_to_port_like_obj',
                       autospec=True)
    @mock.patch.object(common, 'plug_port_to_tenant_network', autospec=True)
    @mock.patch.object(common, 'get_free_port_like_object', autospec=True)
    @mock.patch.object(neutron_common, 'get_client', autospec=True)
    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    @mock.patch.object(neutron_common, 'get_physnets_by_port_uuid',
                       autospec=True)
    def test_vif_attach_active_node_failure(self, mock_gpbpi, mock_upa,
                                            mock_client, mock_gfp, mock_plug,
                                            mock_save):
        self.node.provision_state = states.ACTIVE
        self.node.save()
        vif = {'id': "fake_vif_id"}
        mock_gfp.return_value = self.port
        mock_plug.side_effect = exception.NetworkError
        with task_manager.acquire(self.context, self.node.id) as task:
            self.assertRaises(exception.NetworkError,
                              self.interface.vif_attach, task, vif)
            mock_client.assert_called_once_with(context=task.context)
            mock_upa.assert_called_once_with(
                "fake_vif_id", self.port.address, context=task.context)
        mock_gpbpi.assert_called_once_with(mock_client.return_value,
                                           'fake_vif_id')
        mock_gfp.assert_called_once_with(task, 'fake_vif_id',
                                         mock_gpbpi.return_value,
                                         {'id': 'fake_vif_id'})
        mock_save.assert_called_once_with(self.port, "fake_vif_id")
        mock_plug.assert_called_once_with(task, self.port, mock.ANY)

    @mock.patch.object(neutron_common, 'update_neutron_port',
                       autospec=True)
    @mock.patch.object(common.VIFPortIDMixin, '_save_vif_to_port_like_obj',
                       autospec=True)
    @mock.patch.object(common, 'get_free_port_like_object', autospec=True)
    @mock.patch.object(neutron_common, 'get_client', autospec=True)
    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    @mock.patch.object(neutron_common, 'get_physnets_by_port_uuid',
                       autospec=True)
    def test_vif_attach_portgroup_no_address(self, mock_gpbpi, mock_upa,
                                             mock_client, mock_gfp, mock_save,
                                             mock_port_update):
        pg = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id, address=None)
        mock_gfp.return_value = pg
        vif = {'id': "fake_vif_id"}
        with task_manager.acquire(self.context, self.node.id) as task:
            self.interface.vif_attach(task, vif)
            mock_client.assert_called_once_with(context=task.context)
            mock_port_update.assert_has_calls([
                mock.call(None, 'fake_vif_id',
                          {'binding:host_id': '', 'binding:profile': {}},
                          None),
                mock.call(None, 'fake_vif_id',
                          {'mac_address': None}, None),
                mock.call(task.context, mock.ANY,
                          {'binding:vnic_type': 'baremetal',
                           'binding:host_id': mock.ANY,
                           'binding:profile': {}},
                          client=None)
            ])

        mock_gpbpi.assert_called_once_with(mock_client.return_value,
                                           'fake_vif_id')
        mock_gfp.assert_called_once_with(task, 'fake_vif_id',
                                         mock_gpbpi.return_value,
                                         {'id': 'fake_vif_id'})
        self.assertFalse(mock_client.return_value.show_port.called)
        self.assertFalse(mock_upa.called)
        mock_save.assert_called_once_with(pg, "fake_vif_id")

    @mock.patch.object(common.VIFPortIDMixin, '_save_vif_to_port_like_obj',
                       autospec=True)
    @mock.patch.object(neutron_common, 'get_client', autospec=True)
    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    @mock.patch.object(neutron_common, 'get_physnets_by_port_uuid',
                       autospec=True)
    def test_vif_attach_update_port_exception(self, mock_gpbpi, mock_upa,
                                              mock_client, mock_save):
        self.port.internal_info = {}
        self.port.physical_network = 'physnet1'
        self.port.save()
        vif = {'id': "fake_vif_id"}
        mock_gpbpi.return_value = {'physnet1'}
        mock_upa.side_effect = (
            exception.FailedToUpdateMacOnPort(port_id='fake'))
        with task_manager.acquire(self.context, self.node.id) as task:
            self.assertRaisesRegex(
                exception.NetworkError, "can not update Neutron port",
                self.interface.vif_attach, task, vif)
            mock_client.assert_called_once_with(context=task.context)
        mock_gpbpi.assert_called_once_with(mock_client.return_value,
                                           'fake_vif_id')
        self.assertFalse(mock_save.called)

    @mock.patch.object(common.VIFPortIDMixin, '_save_vif_to_port_like_obj',
                       autospec=True)
    @mock.patch.object(common, 'get_free_port_like_object', autospec=True)
    @mock.patch.object(neutron_common, 'get_client', autospec=True)
    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    @mock.patch.object(neutron_common, 'get_physnets_by_port_uuid',
                       autospec=True)
    def test_vif_attach_portgroup_physnet_inconsistent(self, mock_gpbpi,
                                                       mock_upa, mock_client,
                                                       mock_gfp, mock_save):
        self.port.physical_network = 'physnet1'
        self.port.save()
        vif = {'id': "fake_vif_id"}
        mock_gpbpi.return_value = {'anyphysnet'}
        mock_gfp.side_effect = exception.PortgroupPhysnetInconsistent(
            portgroup='fake-portgroup-id', physical_networks='physnet1')
        with task_manager.acquire(self.context, self.node.id) as task:
            self.assertRaises(
                exception.PortgroupPhysnetInconsistent,
                self.interface.vif_attach, task, vif)
            mock_client.assert_called_once_with(context=task.context)
        mock_gpbpi.assert_called_once_with(mock_client.return_value,
                                           'fake_vif_id')
        self.assertFalse(mock_upa.called)
        self.assertFalse(mock_save.called)

    @mock.patch.object(common.VIFPortIDMixin, '_save_vif_to_port_like_obj',
                       autospec=True)
    @mock.patch.object(common, 'get_free_port_like_object', autospec=True)
    @mock.patch.object(neutron_common, 'get_client', autospec=True)
    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    @mock.patch.object(neutron_common, 'get_physnets_by_port_uuid',
                       autospec=True)
    def test_vif_attach_trait_based_networking_enabled(self, mock_gpbpi,
                                                       mock_upa, mock_client,
                                                       mock_gfp, mock_save):
        # Trait will match TBN trait defined below.
        self.node.set_instance_info('traits', ['CUSTOM_TRAIT_NAME'])
        self.node.save()

        # Vendor matches filter expression below.
        self.port.vendor = 'fake_vendor'
        self.port.internal_info = {}
        self.port.save()

        vif_info = {'id': 'fake_vif_id'}

        contents = ("CUSTOM_TRAIT_NAME:\n"
                    "  actions:\n"
                    "    - action: attach_port\n"
                    "      filter: port.vendor == 'fake_vendor'\n")
        self.tmpdir = tempfile.TemporaryDirectory()

        with tempfile.NamedTemporaryFile(mode='w', dir=self.tmpdir.name,
                                         delete=False) as tmpfile:
            tmpfile.write(contents)
            tmpfile.close()

            # Enable TBN through conductor configuration
            self.config(enable_trait_based_networking=True, group='conductor')
            self.config(trait_based_networking_config_file=tmpfile.name,
                        group='conductor')

            with task_manager.acquire(self.context, self.node.id) as task:
                self.interface.vif_attach(task, vif_info)

            self.assertFalse(mock_gpbpi.called)
            self.assertFalse(mock_gfp.called)

            self.assertTrue(mock_save.called)

            # Make sure save was called with the expected parameters.
            self.assertEqual(mock_save.call_args_list[0][0][0].id,
                             self.port.id)
            self.assertEqual(mock_save.call_args_list[0][0][1], vif_info['id'])

    @mock.patch.object(common.VIFPortIDMixin, '_save_vif_to_port_like_obj',
                       autospec=True)
    @mock.patch.object(common, 'get_free_port_like_object', autospec=True)
    @mock.patch.object(neutron_common, 'get_client', autospec=True)
    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    @mock.patch.object(neutron_common, 'get_physnets_by_port_uuid',
                       autospec=True)
    def test_vif_attach_trait_based_networking_default_trait_applies(
            self, mock_gpbpi, mock_upa, mock_client, mock_gfp, mock_save):
        # No trait will match TBN trait defined below.
        self.node.set_instance_info('traits', [])
        self.node.save()

        # Vendor matches filter expression below.
        self.port.vendor = 'bad_vendor'
        self.port.internal_info = {}
        self.port.save()

        vif_info = {'id': 'fake_vif_id'}

        contents = ("CUSTOM_TRAIT_NAME:\n"
                    "  actions:\n"
                    "    - action: attach_port\n"
                    "      filter: port.vendor == 'fake_vendor'\n")
        self.tmpdir = tempfile.TemporaryDirectory()

        with tempfile.NamedTemporaryFile(mode='w', dir=self.tmpdir.name,
                                         delete=False) as tmpfile:
            tmpfile.write(contents)
            tmpfile.close()

            # Enable TBN through conductor configuration
            self.config(enable_trait_based_networking=True, group='conductor')
            self.config(trait_based_networking_config_file=tmpfile.name,
                        group='conductor')

            with task_manager.acquire(self.context, self.node.id) as task:
                self.interface.vif_attach(task, vif_info)

            self.assertFalse(mock_gpbpi.called)
            self.assertFalse(mock_gfp.called)

            self.assertTrue(mock_save.called)

            # Make sure save was called with the expected parameters.
            self.assertEqual(mock_save.call_args_list[0][0][0].id,
                             self.port.id)
            self.assertEqual(mock_save.call_args_list[0][0][1], vif_info['id'])



    @mock.patch.object(common.VIFPortIDMixin, '_save_vif_to_port_like_obj',
                       autospec=True)
    @mock.patch.object(common, 'get_free_port_like_object', autospec=True)
    @mock.patch.object(neutron_common, 'get_client', autospec=True)
    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    @mock.patch.object(neutron_common, 'get_physnets_by_port_uuid',
                       autospec=True)
    def test_vif_attach_trait_based_networking_enabled_no_match_raises(
            self, mock_gpbpi, mock_upa, mock_client, mock_gfp, mock_save):
        # Trait will match TBN trait defined below.
        self.node.set_instance_info('traits', ['CUSTOM_TRAIT_NAME'])
        self.node.save()

        # But port vendor does NOT match filter expression below.
        self.port.vendor = 'bad_vendor'
        self.port.internal_info = {}
        self.port.save()

        vif_info = {'id': 'fake_vif_id'}

        contents = ("CUSTOM_TRAIT_NAME:\n"
                    "  actions:\n"
                    "    - action: attach_port\n"
                    "      filter: port.vendor == 'fake_vendor'\n")
        self.tmpdir = tempfile.TemporaryDirectory()

        with tempfile.NamedTemporaryFile(mode='w', dir=self.tmpdir.name,
                                         delete=False) as tmpfile:
            tmpfile.write(contents)
            tmpfile.close()

            # Enable TBN through conductor configuration
            self.config(enable_trait_based_networking=True, group='conductor')
            self.config(trait_based_networking_config_file=tmpfile.name,
                        group='conductor')

            with task_manager.acquire(self.context, self.node.id) as task:
                self.assertRaises(exception.TBNNoPortMatchesFound ,
                                  self.interface.vif_attach,
                                  task, vif_info)

            self.assertFalse(mock_gpbpi.called)
            self.assertFalse(mock_gfp.called)
            self.assertFalse(mock_save.called)

    @mock.patch.object(common.VIFPortIDMixin, '_save_vif_to_port_like_obj',
                       autospec=True)
    @mock.patch.object(common, 'get_free_port_like_object', autospec=True)
    @mock.patch.object(neutron_common, 'get_client', autospec=True)
    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    @mock.patch.object(neutron_common, 'get_physnets_by_port_uuid',
                       autospec=True)
    def test_vif_attach_trait_based_networking_enabled_multiple_port_matches(
            self, mock_gpbpi, mock_upa, mock_client, mock_gfp, mock_save):
        # Trait will match TBN trait defined below.
        self.node.set_instance_info('traits', ['CUSTOM_TRAIT_NAME'])
        self.node.save()

        # Vendor matches filter expression below.
        self.port.vendor = 'fake_vendor'
        self.port.internal_info = {}
        self.port.save()

        second_port = obj_utils.create_test_port(
                self.context, node_id=self.node.id,
                uuid=uuidutils.generate_uuid(),
                address='52:54:00:cf:2d:33',
                vendor='fake_vendor',
                internal_info={})
        second_port.save()

        vif_info = {'id': 'fake_vif_id'}

        contents = ("CUSTOM_TRAIT_NAME:\n"
                    "  actions:\n"
                    "    - action: attach_port\n"
                    "      filter: port.vendor == 'fake_vendor'\n"
                    "      min_count: 2")
        self.tmpdir = tempfile.TemporaryDirectory()

        with tempfile.NamedTemporaryFile(mode='w', dir=self.tmpdir.name,
                                         delete=False) as tmpfile:
            tmpfile.write(contents)
            tmpfile.close()

            # Enable TBN through conductor configuration
            self.config(enable_trait_based_networking=True, group='conductor')
            self.config(trait_based_networking_config_file=tmpfile.name,
                        group='conductor')

            with task_manager.acquire(self.context, self.node.id) as task:
                self.interface.vif_attach(task, vif_info)

            self.assertFalse(mock_gpbpi.called)
            self.assertFalse(mock_gfp.called)

            self.assertTrue(mock_save.called)

            # Should have been called twice.
            self.assertEqual(len(mock_save.call_args_list), 2)

            # Make sure save was called with the expected parameters.
            # Both ports should have been saved/attached.
            self.assertEqual(mock_save.call_args_list[1][0][0].id,
                             self.port.id)
            self.assertEqual(mock_save.call_args_list[1][0][1], vif_info['id'])

            self.assertEqual(mock_save.call_args_list[0][0][0].id,
                             second_port.id)
            self.assertEqual(mock_save.call_args_list[0][0][1], vif_info['id'])

    @mock.patch.object(common.VIFPortIDMixin, '_save_vif_to_port_like_obj',
                       autospec=True)
    @mock.patch.object(common, 'get_free_port_like_object', autospec=True)
    @mock.patch.object(neutron_common, 'get_client', autospec=True)
    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    @mock.patch.object(neutron_common, 'get_physnets_by_port_uuid',
                       autospec=True)
    def test_vif_attach_trait_based_networking_enabled_handles_no_matches(
            self, mock_gpbpi, mock_upa, mock_client, mock_gfp, mock_save):
        # Trait will match TBN trait defined below.
        self.node.set_instance_info('traits', ['CUSTOM_TRAIT_NAME'])
        self.node.save()

        # Vendor matches filter expression below.
        self.port.vendor = 'fake_vendor'
        self.port.internal_info = {}
        self.port.save()

        vif_info = {'id': 'fake_vif_id'}

        # Once action will generate an attach action,
        # the other will generate a NoMatch action.
        # Make sure things don't blow up.
        contents = ("CUSTOM_TRAIT_NAME:\n"
                    "  actions:\n"
                    "    - action: attach_port\n"
                    "      filter: port.vendor == 'other_vendor'\n"
                    "    - action: attach_port\n"
                    "      filter: port.vendor == 'fake_vendor'\n")
        self.tmpdir = tempfile.TemporaryDirectory()

        with tempfile.NamedTemporaryFile(mode='w', dir=self.tmpdir.name,
                                         delete=False) as tmpfile:
            tmpfile.write(contents)
            tmpfile.close()

            # Enable TBN through conductor configuration
            self.config(enable_trait_based_networking=True, group='conductor')
            self.config(trait_based_networking_config_file=tmpfile.name,
                        group='conductor')

            with task_manager.acquire(self.context, self.node.id) as task:
                self.interface.vif_attach(task, vif_info)

            self.assertFalse(mock_gpbpi.called)
            self.assertFalse(mock_gfp.called)

            self.assertTrue(mock_save.called)

            # Should have been called once.
            self.assertEqual(len(mock_save.call_args_list), 1)

            # Make sure save was called with the expected parameters.
            # Both ports should have been saved/attached.
            self.assertEqual(mock_save.call_args_list[0][0][0].id,
                             self.port.id)
            self.assertEqual(mock_save.call_args_list[0][0][1], vif_info['id'])

    @mock.patch.object(common.VIFPortIDMixin, '_save_vif_to_port_like_obj',
                       autospec=True)
    @mock.patch.object(common, 'get_free_port_like_object', autospec=True)
    @mock.patch.object(neutron_common, 'get_client', autospec=True)
    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    @mock.patch.object(neutron_common, 'get_physnets_by_port_uuid',
                       autospec=True)
    def test_vif_attach_trait_based_networking_enabled_dynamic_portgroup(
            self, mock_gpbpi, mock_upa, mock_client, mock_gfp, mock_save):
        # Trait will match TBN trait defined below.
        self.node.set_instance_info('traits', ['CUSTOM_TRAIT_NAME'])
        self.node.save()

        # Vendor matches filter expression below.
        self.port.vendor = 'dynamic'
        self.port.internal_info = {}
        self.port.available_for_dynamic_portgroup = True
        self.port.save()

        second_port = obj_utils.create_test_port(
                self.context, node_id=self.node.id,
                uuid=uuidutils.generate_uuid(),
                address='52:54:00:cf:2d:33',
                vendor='dynamic',
                internal_info={},
                available_for_dynamic_portgroup=True)
        second_port.save()

        vif_info = {'id': 'fake_vif_id'}

        # This should create a dynamic portgroup with the two existing test
        # ports and attach it to the available network.
        contents = ("CUSTOM_TRAIT_NAME:\n"
                    "  actions:\n"
                    "    - action: group_and_attach_ports\n"
                    "      filter: port.vendor == 'dynamic' \n"
                    "      min_count: 2\n")
        self.tmpdir = tempfile.TemporaryDirectory()

        with tempfile.NamedTemporaryFile(mode='w', dir=self.tmpdir.name,
                                         delete=False) as tmpfile:
            tmpfile.write(contents)
            tmpfile.close()

            # Enable TBN through conductor configuration
            self.config(enable_trait_based_networking=True, group='conductor')
            self.config(trait_based_networking_config_file=tmpfile.name,
                        group='conductor')

            with task_manager.acquire(self.context, self.node.id) as task:
                self.interface.vif_attach(task, vif_info)

            self.assertFalse(mock_gpbpi.called)
            self.assertFalse(mock_gfp.called)

            # Make sure the new portgroup was attached to the network.
            self.assertTrue(mock_save.called)
            self.assertEqual(len(mock_save.call_args_list), 1)

            saved_pg = mock_save.call_args_list[0][0][0]
            self.assertTrue(saved_pg.name.startswith('tbn_dyn_portgroup_'))
            self.assertEqual(mock_save.call_args_list[0][0][1], vif_info['id'])

            # Make sure the selected ports are now part of the dynamic
            # portgroup.
            self.port.refresh()
            self.assertEqual(self.port.portgroup_id, saved_pg.id)
            second_port.refresh()
            self.assertEqual(second_port.portgroup_id, saved_pg.id)


    @mock.patch.object(common.VIFPortIDMixin, '_clear_vif_from_port_like_obj',
                       autospec=True)
    @mock.patch.object(neutron_common, 'unbind_neutron_port', autospec=True)
    @mock.patch.object(common.VIFPortIDMixin, '_get_port_like_obj_by_vif_id',
                       autospec=True)
    def test_vif_detach(self, mock_get, mock_unp, mock_clear):
        mock_get.return_value = self.port
        with task_manager.acquire(self.context, self.node.id) as task:
            self.interface.vif_detach(task, 'fake_vif_id')
        mock_get.assert_called_once_with(self.interface, task, 'fake_vif_id')
        self.assertTrue(mock_unp.called)
        mock_clear.assert_called_once_with(self.port)

    @mock.patch.object(common.VIFPortIDMixin, '_clear_vif_from_port_like_obj',
                       autospec=True)
    @mock.patch.object(neutron_common, 'unbind_neutron_port', autospec=True)
    @mock.patch.object(common.VIFPortIDMixin, '_get_port_like_obj_by_vif_id',
                       autospec=True)
    def test_vif_detach_portgroup(self, mock_get, mock_unp, mock_clear):
        pg = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id)
        mock_get.return_value = pg
        with task_manager.acquire(self.context, self.node.id) as task:
            self.interface.vif_detach(task, 'fake_vif_id')
        mock_get.assert_called_once_with(self.interface, task, 'fake_vif_id')
        self.assertTrue(mock_unp.called)
        mock_clear.assert_called_once_with(pg)

    @mock.patch.object(common.VIFPortIDMixin, '_clear_vif_from_port_like_obj',
                       autospec=True)
    @mock.patch.object(neutron_common, 'unbind_neutron_port', autospec=True)
    @mock.patch.object(common.VIFPortIDMixin, '_get_port_like_obj_by_vif_id',
                       autospec=True)
    def test_vif_detach_not_attached(self, mock_get, mock_unp, mock_clear):
        mock_get.side_effect = exception.VifNotAttached(vif='fake-vif',
                                                        node='fake-node')
        with task_manager.acquire(self.context, self.node.id) as task:
            self.assertRaisesRegex(
                exception.VifNotAttached, "it is not attached to it.",
                self.interface.vif_detach, task, 'fake_vif_id')
        mock_get.assert_called_once_with(self.interface, task, 'fake_vif_id')
        self.assertFalse(mock_unp.called)
        self.assertFalse(mock_clear.called)

    @mock.patch.object(common.VIFPortIDMixin, '_clear_vif_from_port_like_obj',
                       autospec=True)
    @mock.patch.object(neutron_common, 'unbind_neutron_port', autospec=True)
    @mock.patch.object(common.VIFPortIDMixin, '_get_port_like_obj_by_vif_id',
                       autospec=True)
    def test_vif_detach_active_node(self, mock_get, mock_unp, mock_clear):
        self.node.provision_state = states.ACTIVE
        self.node.save()
        mock_get.return_value = self.port
        with task_manager.acquire(self.context, self.node.id) as task:
            self.interface.vif_detach(task, 'fake_vif_id')
            mock_unp.assert_called_once_with('fake_vif_id',
                                             context=task.context)
        mock_get.assert_called_once_with(self.interface, task, 'fake_vif_id')
        mock_clear.assert_called_once_with(self.port)

    @mock.patch.object(common.VIFPortIDMixin, '_clear_vif_from_port_like_obj',
                       autospec=True)
    @mock.patch.object(neutron_common, 'unbind_neutron_port', autospec=True)
    @mock.patch.object(common.VIFPortIDMixin, '_get_port_like_obj_by_vif_id',
                       autospec=True)
    def test_vif_detach_deleting_node(self, mock_get, mock_unp, mock_clear):
        self.node.provision_state = states.DELETING
        self.node.save()
        mock_get.return_value = self.port
        with task_manager.acquire(self.context, self.node.id) as task:
            self.interface.vif_detach(task, 'fake_vif_id')
            mock_unp.assert_called_once_with('fake_vif_id',
                                             context=task.context)
        mock_get.assert_called_once_with(self.interface, task, 'fake_vif_id')
        mock_clear.assert_called_once_with(self.port)

    @mock.patch.object(common.VIFPortIDMixin, '_clear_vif_from_port_like_obj',
                       autospec=True)
    @mock.patch.object(neutron_common, 'unbind_neutron_port', autospec=True)
    @mock.patch.object(common.VIFPortIDMixin, '_get_port_like_obj_by_vif_id',
                       autospec=True)
    def test_vif_detach_active_node_failure(self, mock_get, mock_unp,
                                            mock_clear):
        self.node.provision_state = states.ACTIVE
        self.node.save()
        mock_get.return_value = self.port
        mock_unp.side_effect = exception.NetworkError
        with task_manager.acquire(self.context, self.node.id) as task:
            self.assertRaises(exception.NetworkError,
                              self.interface.vif_detach, task, 'fake_vif_id')
            mock_unp.assert_called_once_with('fake_vif_id',
                                             context=task.context)
        mock_get.assert_called_once_with(self.interface, task, 'fake_vif_id')
        mock_clear.assert_called_once_with(self.port)

    @mock.patch.object(common.VIFPortIDMixin, '_clear_vif_from_port_like_obj',
                       autospec=True)
    @mock.patch.object(neutron_common, 'unbind_neutron_port', autospec=True)
    @mock.patch.object(common.VIFPortIDMixin, '_get_port_like_obj_by_vif_id',
                       autospec=True)
    def test_vif_detach_dynamic_portgroup(self, mock_get, mock_unp,
                                          mock_clear):
        # Create a dynamic portgroup.
        pg = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id, dynamic_portgroup=True)

        # Add some ports to it.
        self.port.portgroup_id = pg.id
        second_port = obj_utils.create_test_port(
                self.context, node_id=self.node.id,
                uuid=uuidutils.generate_uuid(),
                address='52:54:00:cf:2d:33',
                vendor='fake_vendor',
                internal_info={},
                portgroup_id=pg.id)
        second_port.save()

        mock_get.return_value = pg
        with task_manager.acquire(self.context, self.node.id) as task:
            self.interface.vif_detach(task, 'fake_vif_id')
        mock_get.assert_called_once_with(self.interface, task, 'fake_vif_id')
        self.assertTrue(mock_unp.called)
        mock_clear.assert_called_once_with(pg)

        # Make sure ports are no longer part  of the portgroup.
        self.port.refresh()
        self.assertIsNone(self.port.portgroup_id)
        second_port.refresh()
        self.assertIsNone(second_port.portgroup_id)

        # Ensure pg is gone.
        with self.assertRaises(exception.PortgroupNotFound):
            portgroup.Portgroup.get_by_uuid(self.context, pg.uuid)

    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    def test_port_changed_address(self, mac_update_mock):
        new_address = '11:22:33:44:55:bb'
        self.port.address = new_address
        with task_manager.acquire(self.context, self.node.id) as task:
            self.interface.port_changed(task, self.port)
            mac_update_mock.assert_called_once_with(
                self.port.internal_info[common.NetType.TENANT.vif_key],
                new_address,
                context=task.context)

    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    def test_port_changed_address_VIF_MAC_update_fail(self, mac_update_mock):
        new_address = '11:22:33:44:55:bb'
        self.port.address = new_address
        mac_update_mock.side_effect = (
            exception.FailedToUpdateMacOnPort(port_id=self.port.uuid))
        with task_manager.acquire(self.context, self.node.id) as task:
            self.assertRaises(exception.FailedToUpdateMacOnPort,
                              self.interface.port_changed,
                              task, self.port)
            mac_update_mock.assert_called_once_with(
                self.port.internal_info[common.NetType.TENANT.vif_key],
                new_address,
                context=task.context)

    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    def test_port_changed_address_no_vif_id(self, mac_update_mock):
        self.port.internal_info = {}
        self.port.save()
        self.port.address = '11:22:33:44:55:bb'
        with task_manager.acquire(self.context, self.node.id) as task:
            self.interface.port_changed(task, self.port)
            self.assertFalse(mac_update_mock.called)

    @skip("https://bugs.launchpad.net/ironic/+bug/2024994")
    @mock.patch('ironic.dhcp.neutron.NeutronDHCPApi.update_port_dhcp_opts',
                autospec=True)
    def test_port_changed_client_id(self, dhcp_update_mock):
        expected_ii = {common.NetType.TENANT.vif_key: 'fake-id'}
        expected_extra = {'client-id': 'fake2'}
        expected_dhcp_opts = [{'opt_name': '61', 'opt_value': 'fake2'}]
        self.port.extra = expected_extra
        self.port.internal_info = expected_ii
        what_changed_mock = mock.Mock()
        what_changed_mock.return_value = ['extra', 'internal_info']
        self.port.obj_what_changed = what_changed_mock
        with task_manager.acquire(self.context, self.node.id) as task:
            self.interface.port_changed(task, self.port)
            dhcp_update_mock.assert_called_once_with(
                mock.ANY, 'fake-id', expected_dhcp_opts, context=task.context)
        self.assertEqual(2, what_changed_mock.call_count)

    @skip("https://bugs.launchpad.net/ironic/+bug/2024994")
    @mock.patch('ironic.dhcp.neutron.NeutronDHCPApi.update_port_dhcp_opts',
                autospec=True)
    def test_port_changed_extra_add_new_key(self, dhcp_update_mock):
        self.port.internal_info = {common.NetType.TENANT.vif_key: 'fake-id'}
        self.port.save()
        expected_extra = self.port.extra
        expected_extra['foo'] = 'bar'
        self.port.extra = expected_extra
        with task_manager.acquire(self.context, self.node.id) as task:
            self.interface.port_changed(task, self.port)
            self.assertFalse(dhcp_update_mock.called)

    @skip("https://bugs.launchpad.net/ironic/+bug/2024994")
    @mock.patch('ironic.dhcp.neutron.NeutronDHCPApi.update_port_dhcp_opts',
                autospec=True)
    def test_port_changed_client_id_fail(self, dhcp_update_mock):
        port = obj_utils.create_test_port(
            self.context,
            id=1119,
            uuid=uuidutils.generate_uuid(),
            node_id=self.node.id,
            address='51:51:00:cf:2d:33',
            internal_info={
                common.NetType.TENANT.vif_key: uuidutils.generate_uuid()
            },
            extra={'client-id': 'fake1'})

        port.internal_info = {common.NetType.TENANT.vif_key: 'fake-id'}
        port.extra = {'client-id': 'fake3'}
        what_changed_mock = mock.Mock()
        what_changed_mock.return_value = ['extra']
        port.obj_what_changed = what_changed_mock
        dhcp_update_mock.side_effect = (
            exception.FailedToUpdateDHCPOptOnPort(port_id=port.uuid))
        with task_manager.acquire(self.context, self.node.id) as task:
            self.assertRaises(exception.FailedToUpdateDHCPOptOnPort,
                              self.interface.port_changed,
                              task, port)
        self.assertEqual(2, what_changed_mock.call_count)

    @skip("https://bugs.launchpad.net/ironic/+bug/2024994")
    @mock.patch('ironic.dhcp.neutron.NeutronDHCPApi.update_port_dhcp_opts',
                autospec=True)
    def test_port_changed_client_id_no_vif_id(self, dhcp_update_mock):
        self.port.internal_info = {}
        self.port.extra = {'client-id': 'fake1'}
        self.port.save()
        self.port.extra = {'client-id': 'fake2'}
        with task_manager.acquire(self.context, self.node.id) as task:
            self.interface.port_changed(task, self.port)
            self.assertFalse(dhcp_update_mock.called)

    @skip("https://bugs.launchpad.net/ironic/+bug/2024994")
    @mock.patch('ironic.dhcp.neutron.NeutronDHCPApi.update_port_dhcp_opts',
                autospec=True)
    def test_port_changed_message_format_failure(self, dhcp_update_mock):
        pg = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id,
            standalone_ports_supported=False)

        port = obj_utils.create_test_port(
            self.context, node_id=self.node.id,
            uuid=uuidutils.generate_uuid(),
            address="aa:bb:cc:dd:ee:01",
            internal_info={common.NetType.TENANT.vif_key: 'blah'},
            pxe_enabled=False)
        port.portgroup_id = pg.id

        with task_manager.acquire(self.context, self.node.id) as task:
            self.assertRaisesRegex(exception.Conflict,
                                   "VIF blah is attached to the port",
                                   self.interface.port_changed,
                                   task, port)

    def _test_port_changed(self, has_vif=False, in_portgroup=False,
                           pxe_enabled=True, standalone_ports=True,
                           expect_errors=False):
        pg = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id,
            standalone_ports_supported=standalone_ports)

        extra_vif = {common.NetType.TENANT.vif_key: uuidutils.generate_uuid()}
        if has_vif:
            internal_info = extra_vif
            opposite_extra = {}
        else:
            internal_info = {}
            opposite_extra = extra_vif
        opposite_pxe_enabled = not pxe_enabled

        pg_id = None
        if in_portgroup:
            pg_id = pg.id

        ports = []

        # Update only portgroup id on existed port with different
        # combinations of pxe_enabled/vif_port_id
        p1 = obj_utils.create_test_port(self.context, node_id=self.node.id,
                                        uuid=uuidutils.generate_uuid(),
                                        address="aa:bb:cc:dd:ee:01",
                                        internal_info=internal_info,
                                        pxe_enabled=pxe_enabled)
        p1.portgroup_id = pg_id
        ports.append(p1)

        # Update portgroup_id/pxe_enabled/vif_port_id in one request
        p2 = obj_utils.create_test_port(self.context, node_id=self.node.id,
                                        uuid=uuidutils.generate_uuid(),
                                        address="aa:bb:cc:dd:ee:02",
                                        internal_info=opposite_extra,
                                        pxe_enabled=opposite_pxe_enabled)
        p2.internal_info = internal_info
        p2.pxe_enabled = pxe_enabled
        p2.portgroup_id = pg_id
        ports.append(p2)

        # Update portgroup_id and pxe_enabled
        p3 = obj_utils.create_test_port(self.context, node_id=self.node.id,
                                        uuid=uuidutils.generate_uuid(),
                                        address="aa:bb:cc:dd:ee:03",
                                        internal_info=internal_info,
                                        pxe_enabled=opposite_pxe_enabled)
        p3.pxe_enabled = pxe_enabled
        p3.portgroup_id = pg_id
        ports.append(p3)

        # Update portgroup_id and vif_port_id
        p4 = obj_utils.create_test_port(self.context, node_id=self.node.id,
                                        uuid=uuidutils.generate_uuid(),
                                        address="aa:bb:cc:dd:ee:04",
                                        pxe_enabled=pxe_enabled,
                                        internal_info=opposite_extra)
        p4.internal_info = internal_info
        p4.portgroup_id = pg_id
        ports.append(p4)

        for port in ports:
            with task_manager.acquire(self.context, self.node.id) as task:
                if not expect_errors:
                    self.interface.port_changed(task, port)
                else:
                    self.assertRaises(exception.Conflict,
                                      self.interface.port_changed,
                                      task, port)

    def test_port_changed_novif_pxe_noportgroup(self):
        self._test_port_changed(has_vif=False, in_portgroup=False,
                                pxe_enabled=True,
                                expect_errors=False)

    def test_port_changed_novif_nopxe_noportgroup(self):
        self._test_port_changed(has_vif=False, in_portgroup=False,
                                pxe_enabled=False,
                                expect_errors=False)

    def test_port_changed_vif_pxe_noportgroup(self):
        self._test_port_changed(has_vif=True, in_portgroup=False,
                                pxe_enabled=True,
                                expect_errors=False)

    def test_port_changed_vif_nopxe_noportgroup(self):
        self._test_port_changed(has_vif=True, in_portgroup=False,
                                pxe_enabled=False,
                                expect_errors=False)

    def test_port_changed_novif_pxe_portgroup_standalone_ports(self):
        self._test_port_changed(has_vif=False, in_portgroup=True,
                                pxe_enabled=True, standalone_ports=True,
                                expect_errors=False)

    def test_port_changed_novif_pxe_portgroup_nostandalone_ports(self):
        self._test_port_changed(has_vif=False, in_portgroup=True,
                                pxe_enabled=True, standalone_ports=False,
                                expect_errors=True)

    def test_port_changed_novif_nopxe_portgroup_standalone_ports(self):
        self._test_port_changed(has_vif=False, in_portgroup=True,
                                pxe_enabled=False, standalone_ports=True,
                                expect_errors=False)

    def test_port_changed_novif_nopxe_portgroup_nostandalone_ports(self):
        self._test_port_changed(has_vif=False, in_portgroup=True,
                                pxe_enabled=False, standalone_ports=False,
                                expect_errors=False)

    def test_port_changed_vif_pxe_portgroup_standalone_ports(self):
        self._test_port_changed(has_vif=True, in_portgroup=True,
                                pxe_enabled=True, standalone_ports=True,
                                expect_errors=False)

    def test_port_changed_vif_pxe_portgroup_nostandalone_ports(self):
        self._test_port_changed(has_vif=True, in_portgroup=True,
                                pxe_enabled=True, standalone_ports=False,
                                expect_errors=True)

    def test_port_changed_vif_nopxe_portgroup_standalone_ports(self):
        self._test_port_changed(has_vif=True, in_portgroup=True,
                                pxe_enabled=True, standalone_ports=True,
                                expect_errors=False)

    def test_port_changed_vif_nopxe_portgroup_nostandalone_ports(self):
        self._test_port_changed(has_vif=True, in_portgroup=True,
                                pxe_enabled=False, standalone_ports=False,
                                expect_errors=True)

    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    def test_update_portgroup_address(self, mac_update_mock):
        pg = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id,
            internal_info={common.NetType.TENANT.vif_key: 'fake-id'})
        new_address = '11:22:33:44:55:bb'
        pg.address = new_address
        with task_manager.acquire(self.context, self.node.id) as task:
            self.interface.portgroup_changed(task, pg)
            mac_update_mock.assert_called_once_with(
                'fake-id', new_address, context=task.context)

    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    def test_update_portgroup_remove_address(self, mac_update_mock):
        pg = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id,
            internal_info={common.NetType.TENANT.vif_key: 'fake-id'})
        pg.address = None
        with task_manager.acquire(self.context, self.node.id) as task:
            self.interface.portgroup_changed(task, pg)
        self.assertFalse(mac_update_mock.called)

    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    def test_update_portgroup_address_fail(self, mac_update_mock):
        pg = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id,
            internal_info={common.NetType.TENANT.vif_key: 'fake-id'})
        new_address = '11:22:33:44:55:bb'
        pg.address = new_address
        mac_update_mock.side_effect = (
            exception.FailedToUpdateMacOnPort('boom'))
        with task_manager.acquire(self.context, self.node.id) as task:
            self.assertRaises(exception.FailedToUpdateMacOnPort,
                              self.interface.portgroup_changed,
                              task, pg)
            mac_update_mock.assert_called_once_with(
                'fake-id', new_address, context=task.context)

    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    def test_update_portgroup_address_no_vif(self, mac_update_mock):
        pg = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id)
        new_address = '11:22:33:44:55:bb'
        pg.address = new_address
        with task_manager.acquire(self.context, self.node.id) as task:
            self.interface.portgroup_changed(task, pg)
        self.assertEqual(new_address, pg.address)
        self.assertFalse(mac_update_mock.called)

    @mock.patch.object(neutron_common, 'update_port_address', autospec=True)
    def test_update_portgroup_nostandalone_ports_pxe_ports_exc(
            self, mac_update_mock):
        pg = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id)
        internal_info = {common.NetType.TENANT.vif_key: 'foo'}
        obj_utils.create_test_port(
            self.context, node_id=self.node.id,
            internal_info=internal_info,
            pxe_enabled=True, portgroup_id=pg.id,
            address="aa:bb:cc:dd:ee:01",
            uuid=uuidutils.generate_uuid())

        pg.standalone_ports_supported = False
        with task_manager.acquire(self.context, self.node.id) as task:
            self.assertRaisesRegex(exception.Conflict,
                                   "VIF foo is attached to this port",
                                   self.interface.portgroup_changed,
                                   task, pg)

    def _test_update_portgroup(self, has_vif=False, with_ports=False,
                               pxe_enabled=True, standalone_ports=True,
                               expect_errors=False):
        # NOTE(vsaienko) make sure that old values are opposite to new,
        # to guarantee that object.what_changes() returns true.
        old_standalone_ports_supported = not standalone_ports

        pg = obj_utils.create_test_portgroup(
            self.context, node_id=self.node.id,
            standalone_ports_supported=old_standalone_ports_supported)

        if with_ports:
            internal_info = {}
            if has_vif:
                internal_info = {
                    common.NetType.TENANT.vif_key: uuidutils.generate_uuid()
                }

            obj_utils.create_test_port(
                self.context, node_id=self.node.id,
                internal_info=internal_info,
                pxe_enabled=pxe_enabled, portgroup_id=pg.id,
                address="aa:bb:cc:dd:ee:01",
                uuid=uuidutils.generate_uuid())

        pg.standalone_ports_supported = standalone_ports

        with task_manager.acquire(self.context, self.node.id) as task:
            if not expect_errors:
                self.interface.portgroup_changed(task, pg)
            else:
                self.assertRaises(exception.Conflict,
                                  self.interface.portgroup_changed,
                                  task, pg)

    def test_update_portgroup_standalone_ports_noports(self):
        self._test_update_portgroup(with_ports=False, standalone_ports=True,
                                    expect_errors=False)

    def test_update_portgroup_standalone_ports_novif_pxe_ports(self):
        self._test_update_portgroup(with_ports=True, standalone_ports=True,
                                    has_vif=False, pxe_enabled=True,
                                    expect_errors=False)

    def test_update_portgroup_nostandalone_ports_novif_pxe_ports(self):
        self._test_update_portgroup(with_ports=True, standalone_ports=False,
                                    has_vif=False, pxe_enabled=True,
                                    expect_errors=True)

    def test_update_portgroup_nostandalone_ports_novif_nopxe_ports(self):
        self._test_update_portgroup(with_ports=True, standalone_ports=False,
                                    has_vif=False, pxe_enabled=False,
                                    expect_errors=False)

    def test_update_portgroup_standalone_ports_novif_nopxe_ports(self):
        self._test_update_portgroup(with_ports=True, standalone_ports=True,
                                    has_vif=False, pxe_enabled=False,
                                    expect_errors=False)

    def test_update_portgroup_standalone_ports_vif_pxe_ports(self):
        self._test_update_portgroup(with_ports=True, standalone_ports=True,
                                    has_vif=True, pxe_enabled=True,
                                    expect_errors=False)

    def test_update_portgroup_nostandalone_ports_vif_pxe_ports(self):
        self._test_update_portgroup(with_ports=True, standalone_ports=False,
                                    has_vif=True, pxe_enabled=True,
                                    expect_errors=True)

    def test_update_portgroup_standalone_ports_vif_nopxe_ports(self):
        self._test_update_portgroup(with_ports=True, standalone_ports=True,
                                    has_vif=True, pxe_enabled=False,
                                    expect_errors=False)

    def test_update_portgroup_nostandalone_ports_vif_nopxe_ports(self):
        self._test_update_portgroup(with_ports=True, standalone_ports=False,
                                    has_vif=True, pxe_enabled=False,
                                    expect_errors=True)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import threading

from cheroot.ssl import builtin as cheroot_ssl
from cheroot import wsgi
from oslo_concurrency import processutils
from oslo_log import log as logging
from oslo_service import service
from oslo_service import sslutils

from ironic.api import app
from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import utils
from ironic.conf import CONF


LOG = logging.getLogger(__name__)
_MAX_DEFAULT_WORKERS = 4


def validate_cert_paths(cert_file, key_file):
    if cert_file and not os.path.exists(cert_file):
        raise RuntimeError(_("Unable to find cert_file: %s") % cert_file)
    if key_file and not os.path.exists(key_file):
        raise RuntimeError(_("Unable to find key_file: %s") % key_file)

    if not cert_file or not key_file:
        raise RuntimeError(_("When running server in SSL mode, you must "
                             "specify a valid cert_file and key_file "
                             "paths in your configuration file"))


class BaseWSGIService(service.ServiceBase):

    def __init__(self, name, app, conf, use_ssl=None):
        """Initialize, but do not start the WSGI server.

        :param name: The name of the WSGI server given to the loader.
        :param app: WSGI application to run.
        :param conf: Object to load configuration from.
        :param use_ssl: Whether to use TLS on the socker.
        :returns: None
        """
        self.name = name
        self._conf = conf
        if use_ssl is None:
            use_ssl = conf.use_ssl

        socket_mode = None
        bind_addr = (conf.host_ip, conf.port)
        if conf.unix_socket:
            utils.unlink_without_raise(conf.unix_socket)
            bind_addr = conf.unix_socket
            socket_mode = conf.unix_socket_mode

        self.server = wsgi.Server(
            bind_addr=bind_addr,
            wsgi_app=app,
            server_name=name)

        if use_ssl:
            cert_file = getattr(conf, "cert_file", None)
            key_file = getattr(conf, "key_file", None)

            if not (cert_file and key_file):
                LOG.warning(
                    "Falling back to deprecated [ssl] group for TLS "
                    "credentials: the global [ssl] configuration block is "
                    "deprecated and will be removed in 2026.1"
                )

                # Register global SSL config options and validate the
                # existence of configured certificate/private key file paths,
                # when in secure mode.
                sslutils.is_enabled(CONF)
                cert_file = CONF.ssl.cert_file
                key_file = CONF.ssl.key_file

            validate_cert_paths(cert_file, key_file)

            self.server.ssl_adapter = cheroot_ssl.BuiltinSSLAdapter(
                certificate=cert_file,
                private_key=key_file,
            )

        self._unix_socket = conf.unix_socket
        self._socket_mode = socket_mode
        self._thread = None

    def start(self):
        """Start serving this service using loaded configuration.

        :returns: None
        """
        self.server.prepare()

        if self._unix_socket and self._socket_mode is not None:
            os.chmod(self._unix_socket, self._socket_mode)

        self._thread = threading.Thread(
            target=self.server.serve,
            daemon=True
        )

        self._thread.start()

    def stop(self):
        """Stop serving this API.

        :returns: None
        """
        if self.server:
            self.server.stop()
            if self._thread:
                self._thread.join(timeout=2)

        if self._unix_socket:
            utils.unlink_without_raise(self._unix_socket)

    def wait(self):
        """Wait for the service to stop serving this API.

        :returns: None
        """
        if self._thread:
            self._thread.join()

    def reset(self):
        """No server greenpools to resize."""
        pass


class WSGIService(BaseWSGIService):
    """Provides ability to launch ironic API from wsgi app."""

    def __init__(self, name, use_ssl=False):
        """Initialize, but do not start the WSGI server.

        :param name: The name of the WSGI server given to the loader.
        :param use_ssl: Wraps the socket in an SSL context if True.
        :returns: None
        """
        self.app = app.VersionSelectorApplication()
        self.workers = (
            CONF.api.api_workers
            # NOTE(dtantsur): each worker takes a substantial amount of memory,
            # so we don't want to end up with dozens of them.
            or min(processutils.get_worker_count(), _MAX_DEFAULT_WORKERS)
        )
        if self.workers and self.workers < 1:
            raise exception.ConfigInvalid(
                _("api_workers value of %d is invalid, "
                  "must be greater than 0.") % self.workers)

        super().__init__(name, self.app, CONF.api, use_ssl=use_ssl)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Migrate to UTF8MB4 character encoding

Revision ID: c1fd28861bb9
Revises: 9c0446cb6bc3
Create Date: 2026-01-23 10:00:00.000000

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'c1fd28861bb9'
down_revision = '9c0446cb6bc3'


# List of all tables that need to be converted to UTF8MB4
TABLES = [
    'chassis',
    'conductors',
    'conductor_hardware_interfaces',
    'nodes',
    'ports',
    'portgroups',
    'node_tags',
    'volume_connectors',
    'volume_targets',
    'node_traits',
    'bios_settings',
    'allocations',
    'deploy_templates',
    'deploy_template_steps',
    'node_history',
    'node_inventory',
    'firmware_information',
    'runbooks',
    'runbook_steps',
    'inspection_rules',
]

# When MySQL converts tables to UTF8MB4 using CONVERT TO CHARACTER SET,
# it automatically promotes TEXT columns to MEDIUMTEXT to preserve the same
# character capacity (since UTF8MB4 uses 4 bytes per character vs 3 for
# UTF8MB3). We need to explicitly restore these columns to TEXT type.
# This includes both plain Text columns and oslo.db's JsonEncodedDict/List
# types which are stored as TEXT in MySQL.
# Note: Columns with mysql_as_long=True (LONGTEXT) are not affected.
# Format: {table_name: [(column_name, nullable), ...]}
TEXT_COLUMNS = {
    'allocations': [
        ('last_error', True),
        ('traits', True),
        ('candidate_nodes', True),
        ('extra', True),
    ],
    'bios_settings': [
        ('value', True),
        ('allowable_values', True),
    ],
    'chassis': [
        ('extra', True),
    ],
    'conductors': [
        ('drivers', True),
    ],
    'deploy_template_steps': [
        ('args', False),
    ],
    'deploy_templates': [
        ('extra', True),
    ],
    'node_history': [
        ('event', True),
    ],
    'nodes': [
        ('last_error', True),
        ('maintenance_reason', True),
        ('protected_reason', True),
        ('description', True),
        ('retired_reason', True),
        ('properties', True),
        ('driver_info', True),
        ('driver_internal_info', True),
        ('clean_step', True),
        ('deploy_step', True),
        ('raid_config', True),
        ('target_raid_config', True),
        ('extra', True),
        ('network_data', True),
        ('service_step', True),
    ],
    'portgroups': [
        ('extra', True),
        ('internal_info', True),
        ('properties', True),
    ],
    'ports': [
        ('extra', True),
        ('local_link_connection', True),
        ('internal_info', True),
    ],
    'runbook_steps': [
        ('args', False),
    ],
    'runbooks': [
        ('extra', True),
    ],
    'volume_connectors': [
        ('extra', True),
    ],
    'volume_targets': [
        ('properties', True),
        ('extra', True),
    ],
}


def _verify_utf8mb4_conversion(connection):
    """Verify all tables have been converted to utf8mb4.

    Queries information_schema to check that all Ironic tables are using
    utf8mb4 character set. Raises an exception if any table is not converted.
    """
    # Get the database name from the connection
    db_name = connection.execute(
        sa.text("SELECT DATABASE()")
    ).scalar()

    # Check table character sets
    result = connection.execute(sa.text(
        "SELECT TABLE_NAME, TABLE_COLLATION "
        "FROM information_schema.TABLES "
        "WHERE TABLE_SCHEMA = :db_name "
        "AND TABLE_NAME IN :tables "
        "AND (TABLE_COLLATION IS NULL OR TABLE_COLLATION NOT LIKE 'utf8mb4%')"
    ), {"db_name": db_name, "tables": tuple(TABLES)})

    failed_tables = result.fetchall()
    if failed_tables:
        table_list = ", ".join(f"{t[0]} ({t[1]})" for t in failed_tables)
        raise Exception(
            f"UTF8MB4 migration verification failed. "
            f"Tables not using utf8mb4: {table_list}"
        )


def upgrade():
    # This migration only applies to MySQL/MariaDB databases.
    # For other databases (SQLite, PostgreSQL), this is a no-op.
    connection = op.get_bind()
    if connection.dialect.name != 'mysql':
        return

    # Convert each table to UTF8MB4 character set with unicode collation.
    # This requires MySQL 8.0+ or MariaDB 10.3+ which use DYNAMIC row format
    # by default, allowing index key prefixes up to 3072 bytes.
    for table in TABLES:
        op.execute(
            sa.text(
                f"ALTER TABLE {table} CONVERT TO CHARACTER SET utf8mb4 "
                "COLLATE utf8mb4_unicode_ci"
            )
        )

    # Restore TEXT columns that were promoted to MEDIUMTEXT during conversion.
    # This is necessary to keep the schema in sync with the SQLAlchemy models.
    # We batch all column modifications per table into a single ALTER TABLE
    # statement for efficiency (one table rebuild instead of many).
    for table, columns in TEXT_COLUMNS.items():
        modifications = []
        for column, nullable in columns:
            null_str = "NULL" if nullable else "NOT NULL"
            modifications.append(
                f"MODIFY COLUMN {column} TEXT "
                f"CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_ci {null_str}"
            )
        op.execute(sa.text(
            f"ALTER TABLE {table} " + ", ".join(modifications)
        ))

    # Verify the conversion was successful
    _verify_utf8mb4_conversion(connection)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from oslo_log import log as logging

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import states
from ironic.common import utils
from ironic.conductor import utils as cond_utils
from ironic.conf import CONF
from ironic.drivers import base
from ironic.drivers.modules import deploy_utils
from ironic.drivers.modules import inspect_utils

LOG = logging.getLogger(__name__)

# Internal field to mark whether ironic or inspector manages boot for the node
_IRONIC_MANAGES_BOOT = 'inspector_manage_boot'


def tear_down_managed_boot(task, always_power_off=False):
    errors = []
    ironic_manages_boot = utils.pop_node_nested_field(
        task.node, 'driver_internal_info', _IRONIC_MANAGES_BOOT)
    power_off_done = False

    if ironic_manages_boot:
        # First, perform a graceful shutdown BEFORE ejecting virtual media
        # to avoid filesystem corruption errors on the node. The OS needs
        # access to the virtual media to shut down cleanly.
        if (CONF.inspector.power_off
                and not utils.fast_track_enabled(task.node)
                and not task.node.disable_power_off):
            try:
                cond_utils.node_power_action(task, states.SOFT_POWER_OFF)
                power_off_done = True
            except Exception as exc:
                errors.append(_('unable to power off the node: %s') % exc)
                LOG.exception('Unable to power off node %s for inspection',
                              task.node.uuid)

        # Now it's safe to eject virtual media since the OS has shut down
        try:
            task.driver.boot.clean_up_ramdisk(task)
        except Exception as exc:
            errors.append(_('unable to clean up ramdisk boot: %s') % exc)
            LOG.exception('Unable to clean up ramdisk boot for node %s',
                          task.node.uuid)
        try:
            with cond_utils.power_state_for_network_configuration(task):
                task.driver.network.remove_inspection_network(task)
        except Exception as exc:
            errors.append(_('unable to remove inspection ports: %s') % exc)
            LOG.exception('Unable to remove inspection network for node %s',
                          task.node.uuid)

    if ((ironic_manages_boot or always_power_off)
            and CONF.inspector.power_off
            and not utils.fast_track_enabled(task.node)
            and not power_off_done):
        if task.node.disable_power_off:
            LOG.debug('Rebooting node %s instead of powering it off because '
                      'disable_power_off is set to True', task.node.uuid)
            power_state = states.REBOOT
            err_msg = _('unable to reboot the node: %s')
        else:
            power_state = states.POWER_OFF
            err_msg = _('unable to power off the node: %s')

        try:
            cond_utils.node_power_action(task, power_state)
        except Exception as exc:
            errors.append(err_msg % exc)
            LOG.exception(err_msg, task.node.uuid)

    return errors


def inspection_error_handler(task, error, raise_exc=False, clean_up=True):
    if clean_up:
        tear_down_managed_boot(task)

    task.node.last_error = error
    if raise_exc:
        task.node.save()
        raise exception.HardwareInspectionFailure(error=error)
    else:
        task.process_event('fail')


def ironic_manages_boot(task, raise_exc=False):
    """Whether ironic should manage boot for this node."""
    try:
        task.driver.boot.validate_inspection(task)
    except exception.UnsupportedDriverExtension as e:
        LOG.debug('The boot interface %(iface)s of the node %(node)s does '
                  'not support managed boot for in-band inspection or '
                  'the required options are not populated: %(exc)s',
                  {'node': task.node.uuid,
                   'iface': task.node.get_interface('boot'),
                   'exc': e})
        if raise_exc:
            raise
        return False

    try:
        task.driver.network.validate_inspection(task)
    except exception.UnsupportedDriverExtension as e:
        LOG.debug('The network interface %(iface)s of the node %(node)s does '
                  'not support managed boot for in-band inspection or '
                  'the required options are not populated: %(exc)s',
                  {'node': task.node.uuid,
                   'iface': task.node.get_interface('network'),
                   'exc': e})
        if raise_exc:
            raise
        return False

    return True


def prepare_managed_inspection(task, endpoint):
    """Prepare the boot interface for managed inspection."""
    params = dict(
        utils.parse_kernel_params(CONF.inspector.extra_kernel_params),
        **{'ipa-inspection-callback-url': endpoint})
    if CONF.inspector.force_dhcp:
        # Ensure LLDP collection for inspection on all interfaces.
        params.setdefault('ipa-collect-lldp', '1')
    if utils.fast_track_enabled(task.node):
        params['ipa-api-url'] = deploy_utils.get_ironic_api_url()

    if not task.node.disable_power_off:
        cond_utils.node_power_action(task, states.POWER_OFF)
    with cond_utils.power_state_for_network_configuration(task):
        task.driver.network.add_inspection_network(task)
    task.driver.boot.prepare_ramdisk(task, ramdisk_params=params)


class Common(base.InspectInterface):

    def get_properties(self):
        """Return the properties of the interface.

        :returns: dictionary of <property name>:<property description> entries.
        """
        return {}  # no properties

    def validate(self, task):
        """Validate the driver-specific inspection information.

        If invalid, raises an exception; otherwise returns None.

        :param task: a task from TaskManager.
        :raises: UnsupportedDriverExtension
        """
        utils.parse_kernel_params(CONF.inspector.extra_kernel_params)
        if CONF.inspector.require_managed_boot:
            ironic_manages_boot(task, raise_exc=True)

    def inspect_hardware(self, task):
        """Inspect hardware to obtain the hardware properties.

        Results will be checked in a periodic task.

        :param task: a task from TaskManager.
        :returns: states.INSPECTWAIT
        :raises: HardwareInspectionFailure on failure
        """
        try:
            inspect_utils.create_ports_if_not_exist(task)
        except exception.UnsupportedDriverExtension:
            LOG.debug('Pre-creating ports prior to inspection not supported'
                      ' on node %s.', task.node.uuid)

        manage_boot = ironic_manages_boot(
            task, raise_exc=CONF.inspector.require_managed_boot)

        utils.set_node_nested_field(task.node, 'driver_internal_info',
                                    _IRONIC_MANAGES_BOOT, manage_boot)
        # Make this interface work with the Ironic own /continue_inspection
        # endpoint to simplify migration to the new in-band inspection
        # implementation.
        inspect_utils.cache_lookup_addresses(task.node)
        task.node.save()

        LOG.debug('Starting inspection for node %(uuid)s. Booting is '
                  '%(managed)s by ironic',
                  {'uuid': task.node.uuid,
                   'managed': manage_boot})

        if manage_boot:
            try:
                self._start_managed_inspection(task)
            except Exception as exc:
                LOG.exception('Unable to start managed inspection for node '
                              '%(uuid)s: %(err)s',
                              {'uuid': task.node.uuid, 'err': exc})
                error = _('unable to start inspection: %s') % exc
                inspection_error_handler(task, error, raise_exc=True)
        else:
            self._start_unmanaged_inspection(task)
        return states.INSPECTWAIT

    def _power_on_or_reboot(self, task):
        # Handles disable_power_off properly
        next_state = (states.REBOOT if task.node.disable_power_off
                      else states.POWER_ON)
        cond_utils.node_power_action(task, next_state)


def clean_up(task, finish=True, always_power_off=False):
    errors = tear_down_managed_boot(task, always_power_off=always_power_off)
    if errors:
        errors = ', '.join(errors)
        LOG.error('Inspection clean up failed for node %(uuid)s: %(err)s',
                  {'uuid': task.node.uuid, 'err': errors})
        msg = _('Inspection clean up failed: %s') % errors
        inspection_error_handler(task, msg, raise_exc=False, clean_up=False)
    elif finish:
        LOG.info('Inspection finished successfully for node %s',
                 task.node.uuid)
        task.process_event('done')



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Client side of the networking RPC API.
"""

from oslo_log import log
import oslo_messaging as messaging

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common.json_rpc import client as json_rpc
from ironic.common import release_mappings as versions
from ironic.common import rpc
from ironic.conf import CONF
from ironic.networking import utils as networking_utils
from ironic.objects import base as objects_base

LOG = log.getLogger(__name__)


class NetworkingAPI(object):
    """Client side of the networking RPC API.

    API version history:

    |    1.0 - Initial version.
    """

    # NOTE(alegacy): This must be in sync with manager.NetworkingManager's.
    RPC_API_VERSION = "1.0"

    def __init__(self, topic=None):
        super(NetworkingAPI, self).__init__()
        self.topic = topic
        if self.topic is None:
            if networking_utils.rpc_transport() == "json-rpc":
                # Use host_ip and port from the JSON-RPC config for topic
                host_ip = CONF.ironic_networking_json_rpc.host_ip
                port = CONF.ironic_networking_json_rpc.port
                topic_host = f"{host_ip}:{port}"
                self.topic = f"ironic.{topic_host}"
            else:
                self.topic = rpc.NETWORKING_TOPIC

        serializer = objects_base.IronicObjectSerializer()
        release_ver = versions.RELEASE_MAPPING.get(CONF.pin_release_version)
        version_cap = (
            release_ver.get("networking_rpc")
            if release_ver else self.RPC_API_VERSION
        )

        if networking_utils.rpc_transport() == "json-rpc":
            # Use a dedicated configuration group for networking JSON-RPC
            self.client = json_rpc.Client(
                serializer=serializer,
                version_cap=version_cap,
                conf_group="ironic_networking_json_rpc",
            )
            # Keep the original topic for JSON-RPC (needed for host extraction)
        elif networking_utils.rpc_transport() != "none":
            target = messaging.Target(topic=self.topic, version="1.0")
            self.client = rpc.get_client(
                target, version_cap=version_cap, serializer=serializer
            )
        else:
            self.client = None

    def _prepare_call(self, topic, version=None):
        """Prepare an RPC call.

        :param topic: RPC topic to send to.
        :param version: RPC API version to require.
        """
        topic = topic or self.topic

        # A safeguard for the case someone uses rpc_transport=None
        if self.client is None:
            raise exception.ServiceUnavailable(
                _("Cannot use 'none' RPC to connect to networking service")
            )

        # Normal RPC path
        return self.client.prepare(topic=topic, version=version)

    def get_topic(self):
        """Get RPC topic name for the networking service."""
        return self.topic

    def update_port(
        self,
        context,
        switch_id,
        port_name,
        description,
        mode,
        native_vlan,
        allowed_vlans=None,
        default_vlan=None,
        lag_name=None,
        topic=None,
    ):
        """Update a port configuration on a switch.

        :param context: request context.
        :param switch_id: Identifier for the switch.
        :param port_name: Name of the port on the switch.
        :param description: Description to set for the port.
        :param mode: Port mode ('access', 'trunk', or 'hybrid').
        :param native_vlan: VLAN ID to be set on the port.
        :param allowed_vlans: List of allowed VLAN IDs to be added(optional).
        :param default_vlan: VLAN ID to removed from the port(optional).
        :param lag_name: LAG name if port is part of a link aggregation group.
        :param topic: RPC topic. Defaults to self.topic.
        :raises: InvalidParameterValue if validation fails.
        :raises: NetworkError if the network operation fails.
        :returns: Dictionary containing the updated port configuration.
        """
        cctxt = self._prepare_call(topic=topic, version="1.0")
        return cctxt.call(
            context,
            "update_port",
            switch_id=switch_id,
            port_name=port_name,
            description=description,
            mode=mode,
            native_vlan=native_vlan,
            allowed_vlans=allowed_vlans,
            lag_name=lag_name,
            default_vlan=default_vlan,
        )

    def reset_port(
        self,
        context,
        switch_id,
        port_name,
        native_vlan,
        allowed_vlans=None,
        default_vlan=None,
        topic=None,
    ):
        """Reset a network switch port to default configuration.

        :param context: request context.
        :param switch_id: Identifier for the switch.
        :param port_name: Name of the port on the switch.
        :param native_vlan: VLAN ID to be removed from the port.
        :param allowed_vlans: List of allowed VLAN IDs to be removed(optional).
        :param default_vlan: VLAN ID to restore onto the port(optional).
        :param topic: RPC topic. Defaults to self.topic.
        :raises: InvalidParameterValue if validation fails.
        :raises: NetworkError if the network operation fails.
        :returns: Dictionary containing the reset port configuration.
        """
        cctxt = self._prepare_call(topic=topic, version="1.0")
        return cctxt.call(
            context,
            "reset_port",
            switch_id=switch_id,
            port_name=port_name,
            native_vlan=native_vlan,
            allowed_vlans=allowed_vlans,
            default_vlan=default_vlan,
        )

    def get_switches(self, context, topic=None):
        """Get information about all configured switches.

        :param context: request context.
        :param topic: RPC topic. Defaults to self.topic.
        :raises: NetworkError if the network operation fails.
        :returns: Dictionary with switch_id as key and switch_info as value.
        """
        cctxt = self._prepare_call(topic=topic, version="1.0")
        return cctxt.call(context, "get_switches")

    def update_lag(
        self,
        context,
        switch_ids,
        lag_name,
        description,
        mode,
        native_vlan,
        aggregation_mode,
        allowed_vlans=None,
        default_vlan=None,
        topic=None,
    ):
        """Update a link aggregation group (LAG) configuration.

        :param context: request context.
        :param switch_ids: List of switch identifiers.
        :param lag_name: Name of the LAG.
        :param description: Description for the LAG.
        :param mode: LAG mode ('access' or 'trunk').
        :param native_vlan: VLAN ID to be set for the LAG.
        :param aggregation_mode: Aggregation mode (e.g., 'lacp', 'static').
        :param allowed_vlans: List of allowed VLAN IDs to be added (optional).
        :param default_vlan: VLAN ID to removed from the port(optional).
        :param topic: RPC topic. Defaults to self.topic.
        :raises: InvalidParameterValue if validation fails.
        :raises: NetworkError if the network operation fails.
        :returns: Dictionary containing the updated LAG configuration.
        """
        cctxt = self._prepare_call(topic=topic, version="1.0")
        return cctxt.call(
            context,
            "update_lag",
            switch_ids=switch_ids,
            lag_name=lag_name,
            description=description,
            mode=mode,
            native_vlan=native_vlan,
            aggregation_mode=aggregation_mode,
            allowed_vlans=allowed_vlans,
            default_vlan=default_vlan,
        )

    def delete_lag(
        self, context, switch_ids, lag_name, topic=None
    ):
        """Delete a link aggregation group (LAG) configuration.

        :param context: request context.
        :param switch_ids: List of switch identifiers.
        :param lag_name: Name of the LAG to delete.
        :param topic: RPC topic. Defaults to self.topic.
        :raises: InvalidParameterValue if validation fails.
        :raises: NetworkError if the network operation fails.
        :returns: Dictionary containing the deletion status.
        """
        cctxt = self._prepare_call(topic=topic, version="1.0")
        return cctxt.call(
            context,
            "delete_lag",
            switch_ids=switch_ids,
            lag_name=lag_name,
        )



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ironic.drivers.modules.switch.base import BaseTranslator
from ironic.drivers.modules.switch.base import SwitchDriverBase
from ironic.drivers.modules.switch.base import SwitchDriverException
from ironic.drivers.modules.switch.base import SwitchNotFound
from oslo_log import log as logging
from oslo_utils import importutils

from ironic.networking.utils import parse_vlan_ranges

devices = importutils.try_import('networking_generic_switch.devices')
device_utils = importutils.try_import(
    'networking_generic_switch.devices.utils')

VALID_PORT_MODES = ['access', 'trunk', 'hybrid']
VALID_TRUNK_MODES = ['trunk', 'hybrid']

LOG = logging.getLogger(__name__)


class GenericSwitchDriver(SwitchDriverBase):
    """Generic Switch Standalone Driver implementation.

    This driver provides a switch driver implementation that is decoupled from
    the Neutron ML2 interface.  It provides the same access to the underlying
    generic switch device interface but with a non-Neutron specific API
    interface.  It is intended to be used as a driver within Ironic's
    Standalone Networking Service.
    """
    DRIVER_NAME = 'generic-switch'
    DRIVER_VERSION = '1.0'
    SUPPORTED = True

    _devices = []

    def __init__(self, *args, **kwargs):
        """Perform driver initialization.

        Initialize the driver and load all configured switch devices
        using the same devices module as GenericSwitchDriver.
        """
        super(GenericSwitchDriver, self).__init__()
        if devices is None:
            # Import failed, handle gracefully
            raise ImportError("networking_generic_switch not imported")

        self._devices = devices.get_devices()

        if self._devices:
            LOG.info('Devices have been loaded: %s',
                     list(self._devices.keys()))
        else:
            LOG.warning('No devices have been loaded')

    @classmethod
    def get_translator(cls):
        return GenericSwitchTranslator()

    @staticmethod
    def _get_trunk_details(native_vlan, allowed_vlans):
        """Convert allowed_vlans to sub_ports format expected by devices."""
        sub_ports = []
        for vlan in allowed_vlans or []:
            if vlan != native_vlan:
                # Don't include default VLAN in sub_ports
                sub_ports.append({'segmentation_id': vlan})
        if len(sub_ports) > 0:
            return {'sub_ports': sub_ports}
        return None

    @staticmethod
    def _validate_switch_id(switch_id):
        """Validate switch ID parameter.

        :param switch_id: Switch identifier to validate
        :raises: ValueError if switch_id is invalid
        """
        if not switch_id or not isinstance(switch_id, str):
            raise ValueError("switch_id must be a non-empty string")
        if switch_id.isspace():
            raise ValueError("switch_id cannot be only whitespace")

    @staticmethod
    def _validate_port_name(port_name):
        """Validate port ID parameter.

        :param port_name: Port name to validate
        :raises: ValueError if port_id is invalid
        """
        if not port_name or not isinstance(port_name, str):
            raise ValueError("port_name must be a non-empty string")
        if port_name.isspace():
            raise ValueError("port_name cannot be only whitespace")

    def _get_switch(self, switch_id):
        """Lookup a switch device by ID.

        :param switch_id: MAC address or hostname of the switch
        :returns: Switch device object
        :raises: SwitchNotFound if switch is not found
        :raises: ValueError if switch_id is invalid
        """
        self._validate_switch_id(switch_id)

        switch = device_utils.get_switch_device(
            self._devices, switch_info=switch_id,
            ngs_mac_address=switch_id)

        if not switch:
            raise SwitchNotFound(switch_id)
        return switch

    @staticmethod
    def _validate_port_mode(mode, allowed_vlans):
        """Validate port mode and required parameters.

        :param mode: Port mode to validate
        :param allowed_vlans: List of allowed VLANs
        :raises: ValueError if mode is invalid or required parameters missing
        """
        if mode not in VALID_PORT_MODES:
            raise ValueError(
                f"Invalid mode '{mode}'. Must be one of: {VALID_PORT_MODES}")

        if mode in VALID_TRUNK_MODES:
            if len(allowed_vlans or []) == 0:
                raise ValueError(
                    f"allowed_vlans parameter cannot be empty or missing "
                    f"when mode is '{mode}'")

    @staticmethod
    def _validate_trunk_support(switch, switch_id, mode):
        """Validate that switch supports trunk operations if needed.

        :param switch: Switch device object
        :param switch_id: Switch identifier for error messages
        :param mode: Port mode
        :raises: SwitchDriverException if trunk mode requested but not
          supported
        """
        if mode in VALID_TRUNK_MODES:
            if not switch.support_trunk_on_ports:
                raise SwitchDriverException(
                    f"Switch {switch_id} does not support trunk ports")

    def update_port(self, switch_id, port_name, description, mode, native_vlan,
                    allowed_vlans=None, default_vlan=None,
                    lag_name=None, **kwargs):
        """Update a port configuration on a switch.

        :param switch_id: Identifier for the switch.
        :param port_name: Name of the port on the switch.
        :param description: Description to set for the port.
        :param mode: Port mode ('access', 'trunk', or 'hybrid').
        :param native_vlan: VLAN ID to be set on the port.
        :param allowed_vlans: List of allowed VLAN IDs to be added(optional).
        :param default_vlan: VLAN ID to removed from the port(optional).
        :param lag_name: LAG name if port is part of a LAG.
        :raises: ValueError on parameter validation errors.
        :raises: SwitchNotFound if the specified switch does not exist.
        :raises: SwitchDriverException on configuration failures.
        """
        # Validate parameters
        self._validate_port_name(port_name)
        self._validate_port_mode(mode, allowed_vlans)

        # Lookup the switch device
        switch = self._get_switch(switch_id)

        # Validate switch capabilities
        self._validate_trunk_support(switch, switch_id, mode)

        LOG.info("Updating port %(port_id)s on "
                 "switch %(switch_id)s with mode %(mode)s, "
                 "native_vlan %(native_vlan)s, "
                 "description '%(description)s', "
                 "allowed_vlans %(allowed_vlans)s, "
                 "default_vlan %(default_vlan)s, ",
                 {'port_id': port_name, 'switch_id': switch_id, 'mode': mode,
                  'native_vlan': native_vlan, 'description': description,
                  'allowed_vlans': allowed_vlans,
                  'default_vlan': default_vlan})

        try:
            # Prepare trunk details for non-access modes
            trunk_details = None
            if mode in VALID_TRUNK_MODES:
                trunk_details = (
                    self._get_trunk_details(native_vlan, allowed_vlans))

            # Configure the port
            switch.plug_port_to_network(
                port_name, native_vlan,
                trunk_details=trunk_details,
                default_vlan=default_vlan)

        except Exception as e:
            LOG.error("Failed to update port "
                      "%(port_id)s on switch %(switch_id)s: %(exc)s",
                      {'port_id': port_name, 'switch_id': switch_id, 'exc': e})
            raise SwitchDriverException(message=str(e))

    def reset_port(self, switch_id, port_name,
                   native_vlan=None, allowed_vlans=None, default_vlan=None):
        """Reset a port configuration on a switch.

        :param switch_id: Identifier for the switch.
        :param port_name: Name of the port on the switch.
        :param native_vlan: VLAN ID to be removed from the port.
        :param allowed_vlans: List of allowed VLAN IDs to be removed(optional).
        :param default_vlan: VLAN ID to restore onto the port(optional).
        :raises: ValueError on parameter validation errors.
        :raises: SwitchNotFound if the specified switch does not exist.
        :raises: SwitchDriverException on configuration failures.
        """
        switch = self._get_switch(switch_id)

        LOG.info("Resetting port %(port_id)s on "
                 "switch %(switch_id)s, removing "
                 "native_vlan %(native_vlan)s, setting "
                 "default_vlan %(default_vlan)s",
                 {'port_id': port_name, 'switch_id': switch_id,
                  'native_vlan': native_vlan,
                  'default_vlan': default_vlan})

        try:
            trunk_details = (
                self._get_trunk_details(native_vlan, allowed_vlans))
            switch.delete_port(port_name, native_vlan,
                               trunk_details=trunk_details,
                               default_vlan=default_vlan)
        except Exception as e:
            LOG.error("Failed to reset port "
                      "%(port_id)s on switch %(switch_id)s: %(exc)s",
                      {'port_id': port_name, 'switch_id': switch_id, 'exc': e})
            raise SwitchDriverException(message=str(e))

    def get_switch_info(self, switch_id):
        """Get information about a switch.

        :param switch_id: MAC address or hostname of the switch
        :returns: Dictionary containing switch information
        :raises: SwitchNotFound if switch is not found
        """
        switch = self._get_switch(switch_id)

        return {
            'switch_id': switch_id,
            'device_name': getattr(switch, 'device_name', switch_id),
            'device_type': switch.config.get('device_type', 'unknown'),
            'allowed_vlans': switch.ngs_config.get('ngs_allowed_vlans', None),
        }

    def is_switch_configured(self, switch_id):
        """Check if this driver is configured to manage the specified switch.

        This method should return True if this driver is configured to handle
        the specified switch. This is used by the driver selection logic to
        determine which driver should handle a specific switch.

        :param switch_id: Identifier for the switch to check.
        :returns: True if this driver can manage the specified switch,
                  False otherwise.
        """
        try:
            switch = self._get_switch(switch_id)
        except SwitchNotFound:
            return False

        # Check for essential configuration parameters
        config = getattr(switch, 'config', {})
        device_type = config.get('device_type')

        if not device_type:
            LOG.warning("Switch %s missing device_type configuration",
                        switch_id)
            return False

        # Check if switch has connectivity information
        host_info = config.get('ip') or config.get('host')
        if not host_info:
            LOG.warning("Switch %s missing connection information",
                        switch_id)
            return False

        return True

    def get_switch_ids(self):
        """Get a list of all switch IDs.

        :returns: List of switch IDs
        """
        switch_ids = [switch_id for switch_id, _ in self._devices.items()]
        return switch_ids

    def update_lag(
        self,
        switch_ids,
        lag_name,
        description,
        mode,
        native_vlan,
        aggregation_mode,
        allowed_vlans=None,
        default_vlan=None,
    ):
        """Update LAG configuration across switches.

        :param switch_ids: List of switch identifiers.
        :param lag_name: Name of the LAG.
        :param description: Description for the LAG.
        :param mode: switchport mode ('access' or 'trunk').
        :param native_vlan: VLAN ID to be set for the LAG.
        :param aggregation_mode: Aggregation mode (e.g., 'lacp', 'static').
        :param allowed_vlans: List of allowed VLAN IDs to be added (optional).
        :param default_vlan: VLAN ID to removed from the port(optional).
        :raises: ValueError on parameter validation errors.
        :raises: SwitchNotFound if the specified switch does not exist.
        :raises: SwitchDriverException on configuration failures.
        """
        # Generic switch driver doesn't directly support LAGs
        # This could be extended in the future or by subclasses
        raise SwitchDriverException(
            "LAG operations not supported by generic switch driver")

    def delete_lag(self, switch_ids, lag_name):
        """Delete LAG configuration from switches.

        :param switch_ids: List of switch identifiers.
        :param lag_name: Name of the LAG to delete.
        :raises: ValueError on parameter validation errors.
        :raises: SwitchNotFound if the specified switch does not exist.
        :raises: SwitchDriverException on configuration failures.
        """
        # Generic switch driver doesn't directly support LAGs
        # This could be extended in the future or by subclasses
        raise SwitchDriverException(
            "LAG operations not supported by generic switch driver")


class GenericSwitchTranslator(BaseTranslator):
    """Translates generic config to networking-generic-switch format."""

    def _get_section_name(self, switch_name):
        """Generate section name for networking-generic-switch driver."""
        return f"genericswitch:{switch_name}"

    def _translate_allowed_vlans(self, allowed_vlans):
        """Translate allowed_vlans from Ironic format to NGS format.

        Ironic supports ranges like "100,102-104,106" or ['100', '102-104'].
        networking-generic-switch expects comma-separated individual VLANs
        like "100,102,103,104,106".

        :param allowed_vlans: String or list of VLAN IDs/ranges
        :returns: Comma-separated string of individual VLAN IDs
        """
        # Parse the input (handles both string and list, expands ranges)
        vlan_set = parse_vlan_ranges(allowed_vlans)

        # Handle None (all VLANs allowed) - NGS uses None for this
        if vlan_set is None:
            return None

        # Handle empty set (no VLANs allowed) - NGS uses empty string
        if not vlan_set:
            return ""

        # Convert set to sorted list and join with commas
        vlan_list = sorted(vlan_set)
        return ",".join(str(vlan) for vlan in vlan_list)

    def _translate_switch_config(self, config):
        """Translate from user format to networking-generic-switch format."""
        # networking-generic-switch expects these fields
        translation_map = {
            # Generic field -> driver field
            "address": "ip",
            "device_type": "device_type",
            "username": "username",
            "password": "password",
            "key_file": "key_file",
            "enable_secret": "secret",
            "port": "port",
            "mac_address": "ngs_mac_address",
            "native_vlan": "ngs_port_native_vlan",
            "persist": "ngs_save_configuration",
            "allowed_vlans": "ngs_allowed_vlans",
        }

        # Translate fields
        driver_config = {}
        for user_key, driver_key in translation_map.items():
            if user_key in config and user_key not in [
                "driver_type",
                "allowed_vlans",
            ]:
                driver_config[driver_key] = config[user_key]

        if "allowed_vlans" in config:
            driver_config["ngs_allowed_vlans"] = (
                self._translate_allowed_vlans(config["allowed_vlans"]))

        if "persist" not in config:
            # If the user does not specify a value for persist, set it to False
            # by default since the driver itself will default to True
            driver_config["ngs_save_configuration"] = False

        return driver_config



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import enum

import jsonschema

from ironic.api.schemas.v1 import inspection_rule as schema
from ironic.common import exception
from ironic.common.i18n import _
from ironic.common.inspection_rules import actions
from ironic.common.inspection_rules import operators
from ironic.common.inspection_rules import utils


class InspectionPhase(enum.Enum):
    MAIN = 'main'


# TODO(stephenfin): Everything here can and should be moved to the jsonschema
# schemas, but doing so will change responses.
def validate_rule(rule, built_in=False):
    """Validate an inspection rule using the JSON schema.

    :param rule: The inspection rule to validate.
    :param built_in: Should this rule be treated as built in for validation
    :raises: Invalid if the rule is invalid.
    """
    try:
        jsonschema.validate(rule, schema.create_request_body)
    except jsonschema.ValidationError as e:
        raise exception.Invalid(
            _('Validation failed for inspection rule: %s') % e)

    errors = []

    phase = rule.get('phase', InspectionPhase.MAIN.value)
    if phase not in (p.value for p in InspectionPhase):
        errors.append(
            _('Invalid phase: %(phase)s. Valid phases are: %(valid)s') % {
                'phase': phase, 'valid': ', '.join(
                    [p.value for p in InspectionPhase])
            })

    priority = rule.get('priority', 0)
    if priority < 0 and not built_in:
        errors.append(
            _("Priority cannot be negative for user-defined rules."))
    if priority > 9999 and not built_in:
        errors.append(
            _("Priority must be between 0 and 9999 for user-defined rules."))

    # Additional plugin-specific validation
    for condition in rule.get('conditions', []):
        op, invtd = utils.parse_inverted_operator(
            condition['op'])
        plugin = operators.get_operator(op)
        if not plugin or not callable(plugin):
            errors.append(
                _('Unsupported condition operator: %s') % op)
        try:
            plugin().validate(condition.get('args', {}))
        except ValueError as exc:
            errors.append(_('Invalid parameters for condition operator '
                            '%(op)s: %(error)s') % {'op': op,
                                                    'error': exc})

    for action in rule['actions']:
        plugin = actions.get_action(action['op'])
        if not plugin or not callable(plugin):
            errors.append(_('Unsupported action operator: %s') % action['op'])
        try:
            plugin().validate(action.get('args', {}))
        except ValueError as exc:
            errors.append(_('Invalid parameters for action operator %(op)s: '
                            '%(error)s') % {'op': action['op'], 'error': exc})

    if errors:
        if len(errors) == 1:
            raise exception.Invalid(errors[0])
        else:
            raise exception.Invalid(_('Multiple validation errors occurred: '
                                      '%s') % '; '.join(errors))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add dynamic portgroup fields

Revision ID: 2a3b4c5d6e7f
Revises: c1fd28861bb9
Create Date: 2026-03-04 00:00:00.000000

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '2a3b4c5d6e7f'
down_revision = 'c1fd28861bb9'


def upgrade():
    op.add_column('ports',
                  sa.Column('available_for_dynamic_portgroup',
                            sa.Boolean(),
                            server_default=sa.true(),
                            nullable=False))
    op.add_column('portgroups',
                  sa.Column('dynamic_portgroup',
                            sa.Boolean(),
                            nullable=True))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


index_request_query = {
    'type': 'object',
    'properties': {
        'fields': {
            'type': 'array',
            'items': {
                'enum': [
                    'created_at',
                    'updated_at',
                    'component',
                    'initial_version',
                    'current_version',
                    'last_version_flashed',
                ],
            },
            # OpenAPI-specific properties
            # https://swagger.io/docs/specification/v3_0/serialization/#query-parameters
            'style': 'form',
            'explode': False,
        },
        'detail': {'type': 'boolean'},
    },
    'required': [],
    'additionalProperties': False,
}

_firmware_response_body = {
    'type': 'object',
    'properties': {
        'component': {'type': 'string'},
        'initial_version': {'type': ['string', 'null']},
        'current_version': {'type': ['string', 'null']},
        'last_version_flashed': {'type': ['string', 'null']},
        'created_at': {'type': 'string', 'format': 'date-time'},
        'updated_at': {'type': ['string', 'null'], 'format': 'date-time'},
    },
    # NOTE(adamcarthur) - These are always returned, regardless of the
    # value of the fields parameter
    'required': ['created_at', 'updated_at'],
    'additionalProperties': False,
}

index_response_body = {
    'type': 'object',
    'properties': {
        'firmware': {
            'type': 'array',
            'items': _firmware_response_body,
        },
    },
    'required': ['firmware'],
    'additionalProperties': False,
}
index_response_body['properties'].update({
    'next': {'type': 'string'},
})



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Ramdisk Deploy Interface
"""

from oslo_log import log as logging

from ironic.common import exception
from ironic.common.glance_service import service_utils
from ironic.common.i18n import _
from ironic.common import image_service
from ironic.common import metrics_utils
from ironic.common import states
from ironic.conductor import task_manager
from ironic.conductor import utils as manager_utils
from ironic.drivers import base
from ironic.drivers.modules import agent_base
from ironic.drivers.modules import deploy_utils

LOG = logging.getLogger(__name__)

METRICS = metrics_utils.get_metrics_logger(__name__)


class RamdiskDeploy(agent_base.AgentBaseMixin, agent_base.HeartbeatMixin,
                    base.DeployInterface):

    def get_properties(self):
        return {}

    def supports_deploy(self, task):
        """Check if this interface supports the given deployment.

        Ramdisk deploy is appropriate when
        ``ironic_ramdisk_deploy=True`` is present **and** one of
        the following holds:

        * ``boot_iso`` is set in instance_info, or
        * ``kernel`` and ``ramdisk`` are set in instance_info, or
        * ``image_source`` is a Glance image with ``boot_iso_id``
          property, or
        * ``image_source`` is a Glance image with ``kernel_id``
          and ``ramdisk_id`` properties.

        For instance_info cases the sentinel is looked up in
        ``instance_info``; for Glance image cases it is looked up
        in the image properties.

        :param task: A TaskManager instance containing the node to
            act on.
        :returns: True if ramdisk deploy is appropriate.
        """
        instance_info = task.node.instance_info
        if (instance_info.get('boot_iso')
                and instance_info.get('ironic_ramdisk_deploy')):
            return True
        if (instance_info.get('kernel')
                and instance_info.get('ramdisk')
                and instance_info.get('ironic_ramdisk_deploy')):
            return True
        image_source = instance_info.get('image_source')
        if (image_source
                and service_utils.is_glance_image(image_source)):
            try:
                props = deploy_utils.get_image_properties(
                    task.context, image_source)
                if not props.get('ironic_ramdisk_deploy'):
                    return False
                if props.get('boot_iso_id'):
                    return True
                if (props.get('kernel_id')
                        and props.get('ramdisk_id')):
                    return True
            except Exception:
                LOG.warning(
                    'Failed to query Glance image %s '
                    'for ramdisk deploy detection',
                    image_source)
        return False

    def validate(self, task):
        if 'ramdisk_boot' not in task.driver.boot.capabilities:
            raise exception.InvalidParameterValue(
                message=_('Invalid configuration: The boot interface '
                          'must have the `ramdisk_boot` capability. '
                          'You are using an incompatible boot interface.'))
        task.driver.boot.validate(task)

        # Validate node capabilities
        deploy_utils.validate_capabilities(task.node)

    @METRICS.timer('RamdiskDeploy.deploy')
    @base.deploy_step(priority=100)
    @task_manager.require_exclusive_lock
    def deploy(self, task):
        if ('configdrive' in task.node.instance_info
                and 'ramdisk_boot_configdrive' not in
                task.driver.boot.capabilities):
            # TODO(dtantsur): make it an actual error?
            LOG.warning('A configuration drive is present in the ramdisk '
                        'deployment request of node %(node)s with boot '
                        'interface %(drv)s. The configuration drive will be '
                        'ignored for this deployment.',
                        {'node': task.node,
                         'drv': task.node.get_interface('boot')})
        manager_utils.node_power_action(task, states.POWER_OFF)
        # Tenant networks must enable connectivity to the boot
        # location, as reboot() can otherwise be very problematic.
        # IDEA(TheJulia): Maybe a "trusted environment" mode flag
        # that we otherwise fail validation on for drivers that
        # require explicit security postures.
        with manager_utils.power_state_for_network_configuration(task):
            task.driver.network.configure_tenant_networks(task)

        # calling boot.prepare_instance will also set the node
        # to boot, and update the templates accordingly
        task.driver.boot.prepare_instance(task)

        # Power-on the instance, with PXE prepared, we're done.
        manager_utils.node_power_action(task, states.POWER_ON)
        LOG.info('Deployment setup for node %s done', task.node.uuid)
        return None

    @METRICS.timer('RamdiskDeploy.prepare')
    @task_manager.require_exclusive_lock
    def prepare(self, task):
        node = task.node

        deploy_utils.populate_storage_driver_internal_info(task)
        if node.provision_state == states.DEPLOYING:
            # Ask the network interface to validate itself so
            # we can ensure we are able to proceed.
            task.driver.network.validate(task)

            manager_utils.node_power_action(task, states.POWER_OFF)
            # NOTE(TheJulia): If this was any other interface, we would
            # unconfigure tenant networks, add provisioning networks, etc.
            task.driver.storage.attach_volumes(task)
            # Resolve boot image info from Glance when only
            # image_source is provided (e.g. Nova path).
            self._resolve_image_info_from_glance(task)
        if node.provision_state in (states.ACTIVE, states.UNRESCUING):
            # In the event of takeover or unrescue.
            task.driver.boot.prepare_instance(task)

    def _resolve_image_info_from_glance(self, task):
        """Resolve boot_iso or kernel/ramdisk from Glance image properties.

        When only image_source is set (e.g. Nova deploys), query Glance
        for the image properties and populate instance_info with
        boot_iso, or kernel and ramdisk as appropriate.

        :param task: a TaskManager instance.
        """
        node = task.node
        i_info = node.instance_info
        image_source = i_info.get('image_source')

        if (not image_source
                or i_info.get('boot_iso')
                or i_info.get('kernel')
                or not service_utils.is_glance_image(image_source)):
            return

        try:
            img_service = image_service.get_image_service(
                image_source, context=task.context)
            image_props = img_service.show(image_source)['properties']
        except (exception.GlanceConnectionFailed,
                exception.ImageNotAuthorized,
                exception.ImageNotFound,
                exception.Invalid) as e:
            LOG.warning('Failed to get Glance image properties for '
                        'node %(node)s: %(err)s',
                        {'node': node.uuid, 'err': e})
            return

        if image_props.get('boot_iso_id'):
            i_info['boot_iso'] = str(image_props['boot_iso_id'])
            i_info['original_image_source'] = str(image_source)
        elif (image_props.get('kernel_id')
                and image_props.get('ramdisk_id')):
            i_info['kernel'] = str(image_props['kernel_id'])
            i_info['ramdisk'] = str(image_props['ramdisk_id'])
        else:
            return

        # TODO(JayF): Image metadata was already inspected before
        #             prepare was called on the deploy driver, so
        #             we need to clear out invalid metadata that was
        #             gleaned before we got here. Ideally, we'd
        #             improve ordering such that we never need to do this.
        i_info.pop('image_source', None)
        i_info.pop('image_type', None)
        node.del_driver_internal_info('is_whole_disk_image')

        # NOTE(JayF): The presence of i_info[image_source] is taken as
        #             a sentinel value to mean "direct deploy". It cannot
        #             be left in instance_info.
        i_info['original_image_source'] = str(image_source)
        i_info.pop('image_source', None)
        node.instance_info = i_info
        node.save()



#!/usr/bin/env python3

import hashlib
import json
import os
import socket
import ssl
import sys

from urllib.parse import urlparse

app_name = os.environ.get("APP", "fake")
app_info = json.loads(os.environ.get("APP_INFO"))
verify = app_info.get("verify_ca", True)
print("""# PSM Certificate Override Settings file
# This is a generated file!  Do not edit.
""")

address = app_info.get("address")
if verify or not address:
    sys.exit(0)

try:
    parsed_url = urlparse(address)
    addr = parsed_url.hostname
    port = parsed_url.port or 443

    context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    context.check_hostname = False
    context.verify_mode = ssl.CERT_NONE

    with socket.create_connection((addr, int(port)), timeout=5) as sock:
        with context.wrap_socket(sock, server_hostname=addr) as wrappedSocket:
            der_cert_bin = wrappedSocket.getpeercert(True)

            digest = hashlib.sha256(der_cert_bin).hexdigest()
            formatted_digest = ':'.join(
                a + b for a, b in zip(digest[::2], digest[1::2])).upper()

    print(f"{addr}:{port}:\tOID.2.16.840.1.101.3.4.2.1\t{formatted_digest}\t")
except Exception as e:
    print("# Problem fetching certificate fingerprint.")
    print(f"# {e}")